from typing import Callable, Dict, Iterator, Optional, Tuple

from comb_spec_searcher.strategies import Rule
from comb_spec_searcher.strategies.strategy import (
    DisjointUnionStrategy,
    StrategyFactory,
)
from tilings import GriddedPerm, Tiling
from tilings.algorithms import Sliding

_NO_SYMMETRY, _ROTATE, _REVERSE, _ROTATE_AND_REVERSE = range(4)


def _tiling_identity_map(tiling: Tiling) -> Tiling:
    return tiling


def _gp_identity_map(gp: GriddedPerm) -> GriddedPerm:
    return gp


def _gp_reverse(c: int) -> Callable[[GriddedPerm], GriddedPerm]:
    def _tmp_func(gp: GriddedPerm) -> GriddedPerm:
        return GriddedPerm(
            gp.patt.reverse(), ((c - x - 1, 0) for x, _ in reversed(gp.pos))
        )

    return _tmp_func


def _gp_rotate90(r: int) -> Callable[[GriddedPerm], GriddedPerm]:
    def _tmp_func(gp: GriddedPerm) -> GriddedPerm:
        return GriddedPerm(gp.patt.rotate(), ((y, 0) for _, y in gp.pos))

    return _tmp_func


def _gp_rotate90_inverse(r: int) -> Callable[[GriddedPerm], GriddedPerm]:
    def _tmp_func(gp: GriddedPerm) -> GriddedPerm:
        return GriddedPerm(gp.patt.rotate(-1), ((0, x) for x, _ in gp.pos))

    return _tmp_func


def _gp_rotate90_and_reverse(r: int) -> Callable[[GriddedPerm], GriddedPerm]:
    def _tmp_func(gp: GriddedPerm) -> GriddedPerm:
        return GriddedPerm(
            gp.patt.rotate(1).reverse(), ((r - y - 1, 0) for _, y in gp.pos)
        )

    return _tmp_func


def _gp_rotate90_and_reverse_inverse(r: int) -> Callable[[GriddedPerm], GriddedPerm]:
    def _tmp_func(gp: GriddedPerm) -> GriddedPerm:
        return GriddedPerm(
            gp.patt.reverse().rotate(-1), ((0, r - x - 1) for x, _ in gp.pos)
        )

    return _tmp_func


class _AdditionalMaps:
    def __init__(
        self,
        t_map: Callable[[Tiling], Tiling] = _tiling_identity_map,
        t_inv: Callable[[Tiling], Tiling] = _tiling_identity_map,
        g_map: Callable[[GriddedPerm], GriddedPerm] = _gp_identity_map,
        g_inv: Callable[[GriddedPerm], GriddedPerm] = _gp_identity_map,
    ) -> None:
        self.t_map = t_map
        self.t_inv = t_inv
        self.g_map = g_map
        self.g_inv = g_inv

    @classmethod
    def reverse(cls, c: int) -> "_AdditionalMaps":
        return _AdditionalMaps(
            Tiling.reverse,
            Tiling.reverse,
            _gp_reverse(c),
            _gp_reverse(c),
        )

    @classmethod
    def rotate90(cls, r: int) -> "_AdditionalMaps":
        return _AdditionalMaps(
            Tiling.rotate90,
            Tiling.rotate270,
            _gp_rotate90(r),
            _gp_rotate90_inverse(r),
        )

    @classmethod
    def rotate90_and_reverse(cls, r: int) -> "_AdditionalMaps":
        return _AdditionalMaps(
            Tiling.antidiagonal,
            Tiling.antidiagonal,
            _gp_rotate90_and_reverse(r),
            _gp_rotate90_and_reverse_inverse(r),
        )

    @classmethod
    def enum_to_map(cls, symmetry_type: int, tiling_size: int) -> "_AdditionalMaps":
        if symmetry_type == _NO_SYMMETRY:
            return cls()
        if symmetry_type == _REVERSE:
            return cls.reverse(tiling_size)
        if symmetry_type == _ROTATE:
            return cls.rotate90(tiling_size)
        if symmetry_type == _ROTATE_AND_REVERSE:
            return cls.rotate90_and_reverse(tiling_size)
        raise ValueError("Unknown symmetry type")

    @staticmethod
    def enum_to_str(symmetry_type: int) -> str:
        if symmetry_type == _NO_SYMMETRY:
            return "no symmetry"
        if symmetry_type == _REVERSE:
            return "reverse"
        if symmetry_type == _ROTATE:
            return "rotate"
        if symmetry_type == _ROTATE_AND_REVERSE:
            return "antidiagonal"
        raise ValueError("Unknown symmetry type")


class SlidingStrategy(DisjointUnionStrategy[Tiling, GriddedPerm]):
    """
    A class for a specific slidding strategy. The init gets the index of both
    column you slidding.
    """

    def __init__(self, av_12: int, av_123: int, symmetry_type: int):
        super().__init__(possibly_empty=False)
        self.av_12 = av_12
        self.av_123 = av_123
        self.symmetry_type = symmetry_type

    def decomposition_function(self, tiling: Tiling) -> Tuple[Tiling]:
        """Return the slidded tiling if it slides, otherwise None."""
        maps = _AdditionalMaps.enum_to_map(self.symmetry_type, max(tiling.dimensions))
        sliding = Sliding(maps.t_map(tiling))
        return (maps.t_inv(sliding.slide_column(self.av_12, self.av_123)),)

    def formal_step(self) -> str:
        sym = _AdditionalMaps.enum_to_str(self.symmetry_type)
        return f"slide {self.av_123} through {self.av_12} after applying {sym}"

    def backward_map(
        self,
        tiling: Tiling,
        gps: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        maps = _AdditionalMaps.enum_to_map(self.symmetry_type, max(tiling.dimensions))
        if self.av_12 < self.av_123:
            yield from (
                maps.g_inv(Sliding.slide_gp(maps.g_map(gp), self.av_12, self.av_123))
                for gp in gps
                if gp is not None
            )
        else:
            yield from (
                maps.g_inv(
                    Sliding.slide_gp_inverse(maps.g_map(gp), self.av_123, self.av_12)
                )
                for gp in gps
                if gp is not None
            )

    def forward_map(
        self,
        tiling: Tiling,
        gp: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[GriddedPerm]:
        maps = _AdditionalMaps.enum_to_map(self.symmetry_type, max(tiling.dimensions))
        if self.av_123 < self.av_12:
            return (
                maps.g_inv(Sliding.slide_gp(maps.g_map(gp), self.av_123, self.av_12)),
            )
        return (
            maps.g_inv(
                Sliding.slide_gp_inverse(maps.g_map(gp), self.av_12, self.av_123)
            ),
        )

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if not comb_class.extra_parameters:
            return super().extra_parameters(comb_class, children)
        if children is None:
            children = self.decomposition_function(comb_class)
        child = children[0]
        maps = _AdditionalMaps.enum_to_map(
            self.symmetry_type, max(comb_class.dimensions)
        )
        return (
            {
                comb_class.get_assumption_parameter(
                    ass
                ): child.get_assumption_parameter(
                    Sliding.slide_assumption(
                        ass, self.av_12, self.av_123, maps.g_map, maps.g_inv
                    )
                )
                for ass in comb_class.assumptions
            },
        )

    def to_jsonable(self) -> dict:
        """Return a dictionary form of the strategy."""
        return {
            "class_module": "tilings.strategies.sliding",
            "strategy_class": "SlidingStrategy",
            "av_12": self.av_12,
            "av_123": self.av_123,
            "symmetry_type": self.symmetry_type,
        }

    @classmethod
    def from_dict(cls, d: dict) -> "SlidingStrategy":
        return cls(
            d["av_12"],
            d["av_123"],
            d["symmetry_type"],
        )

    def __str__(self) -> str:
        return self.formal_step()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(av_12={self.av_12}, av_123={self.av_123}, "
            f"symmetry_type={self.symmetry_type})"
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return False
        return all(
            (
                self.av_12 == other.av_12,
                self.av_123 == other.av_123,
                self.symmetry_type == other.symmetry_type,
            )
        )


class SlidingFactory(StrategyFactory[Tiling]):
    """A strategy factory that produces all valid slides for a tiling."""

    def __init__(self, use_symmetries: bool = False):
        super().__init__()
        self.use_symmetries = use_symmetries

    def __call__(self, comb_class: Tiling) -> Iterator[Rule]:
        if comb_class.dimensions[0] > 1 and comb_class.dimensions[1] == 1:
            sliding = Sliding(comb_class)
            for pair in sliding.slidable_pairs():
                child = (sliding.slide_column(*pair),)
                yield SlidingStrategy(*pair, _NO_SYMMETRY)(comb_class, child)
        if self.use_symmetries:
            yield from SlidingFactory._symmetries(comb_class)

    @staticmethod
    def _symmetries(comb_class: Tiling) -> Iterator[Rule]:
        c, r = comb_class.dimensions
        if r == 1 and c > 1:
            tiling = comb_class.reverse()
            sliding = Sliding(tiling)
            for pair in sliding.slidable_pairs():
                maps = _AdditionalMaps.enum_to_map(_REVERSE, c)
                child = (maps.t_inv(sliding.slide_column(*pair)),)
                yield SlidingStrategy(*pair, _REVERSE)(comb_class, child)
        elif r > 1 and c == 1:
            tiling = comb_class.rotate90()
            sliding = Sliding(tiling)
            for pair in sliding.slidable_pairs():
                maps = _AdditionalMaps.enum_to_map(_ROTATE, r)
                child = (maps.t_inv(sliding.slide_column(*pair)),)
                yield SlidingStrategy(*pair, _ROTATE)(comb_class, child)
            tiling = tiling.reverse()
            sliding = Sliding(tiling)
            for pair in sliding.slidable_pairs():
                maps = _AdditionalMaps.enum_to_map(_ROTATE_AND_REVERSE, r)
                child = (maps.t_inv(sliding.slide_column(*pair)),)
                yield SlidingStrategy(*pair, _ROTATE_AND_REVERSE)(comb_class, child)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(use_symmetries={self.use_symmetries})"

    def __str__(self) -> str:
        return f"Sliding{' (with symmetries)' if self.use_symmetries else ''}"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["use_symmetries"] = self.use_symmetries
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "SlidingFactory":
        """
        Return the strategy from the json representation.
        """
        return cls(**d)



from typing import TYPE_CHECKING, Iterable, List, Optional, Union

from logzero import logger

from comb_spec_searcher import StrategyPack
from comb_spec_searcher.strategies import (
    AbstractStrategy,
    Strategy,
    StrategyFactory,
    VerificationStrategy,
)
from permuta import Perm
from permuta.misc import DIR_EAST, DIR_NORTH, DIR_SOUTH, DIR_WEST, DIRS
from tilings import strategies as strat
from tilings.strategies.verification import BasisAwareVerificationStrategy

if TYPE_CHECKING:
    from tilings import Tiling

CSSstrategy = Union[Strategy, StrategyFactory, VerificationStrategy]


class TileScopePack(StrategyPack):
    # Method to add power to a pack
    # Pack are immutable, these methods return a new pack.

    def add_basis(self, basis: Iterable[Perm]) -> "TileScopePack":
        """
        Update the pack to add the basis being run to the verifications strategy
        that needs to be aware of it.
        """
        basis = tuple(basis)
        symmetry = bool(self.symmetries)

        def replace_list(strats):
            """Return a new list with the replaced 1x1 strat."""
            res = []
            for strategy in strats:
                if isinstance(strategy, BasisAwareVerificationStrategy):
                    if strategy.basis:
                        logger.warning("Basis changed in %s", strategy)
                    res.append(strategy.change_basis(basis, symmetry))
                else:
                    res.append(strategy)
            return res

        return self.__class__(
            ver_strats=replace_list(self.ver_strats),
            inferral_strats=replace_list(self.inferral_strats),
            initial_strats=replace_list(self.initial_strats),
            expansion_strats=list(map(replace_list, self.expansion_strats)),
            name=self.name,
            symmetries=self.symmetries,
            iterative=self.iterative,
        )

    def setup_subclass_verification(self, start_tiling: "Tiling") -> "TileScopePack":
        """
        If the subclass verification strategy already has a list of perms to check,
        we leave it alone. Otherwise we:
         - compute the cell basis for each cell
         - consider the permutations avoiding some cell basis
         - add to the list of perms to check any perm that avoids some cell basis and
           has length strictly smaller than the maximum length cell basis element.
        """

        def replace_list(strats):
            """
            Find subclass verification and alter its perms_to_check variable.
            """
            res = []
            for strategy in strats:
                if isinstance(strategy, strat.SubclassVerificationFactory):
                    printed_log = False
                    if strategy.perms_to_check is None:
                        new_perms_to_check = set()
                        cell_bases = set(
                            tuple(obs) for obs, _ in start_tiling.cell_basis().values()
                        )
                        max_length = (
                            max(max(len(b) for b in basis) for basis in cell_bases) - 1
                        )
                        perm_gen = Perm.up_to_length(max_length)
                        for perm in perm_gen:
                            if len(perm) == 0:
                                continue
                            if any(perm.avoids(*basis) for basis in cell_bases):
                                new_perms_to_check.add(perm)
                        res.append(strategy.change_perms(new_perms_to_check))
                        if start_tiling.dimensions == (1, 1):
                            logger.info(
                                "SubclassVerification set up to check the proper "
                                "principal subclasses of Av(%s)",
                                ", ".join(map(str, cell_bases.pop())),
                            )
                            printed_log = True
                    if not printed_log:
                        logger.info(
                            "SubclassVerification set up to check the subclasses: "
                            "Av(%s)",
                            "), Av(".join(map(str, strategy.perms_to_check)),
                        )
                else:
                    res.append(strategy)
            return res

        return self.__class__(
            ver_strats=replace_list(self.ver_strats),
            inferral_strats=replace_list(self.inferral_strats),
            initial_strats=replace_list(self.initial_strats),
            expansion_strats=list(map(replace_list, self.expansion_strats)),
            name=self.name,
            symmetries=self.symmetries,
            iterative=self.iterative,
        )

    def make_tracked(self):
        """Make a fusion pack tracked."""

        def replace_list(strats):
            """Return a new list with the replaced fusion strat."""
            res = []
            for strategy in strats:
                if isinstance(strategy, strat.FusionFactory):
                    res.append(strategy.make_tracked())
                else:
                    res.append(strategy)
            return res

        return (
            self.__class__(
                ver_strats=replace_list(self.ver_strats),
                inferral_strats=replace_list(self.inferral_strats),
                initial_strats=replace_list(self.initial_strats),
                expansion_strats=list(map(replace_list, self.expansion_strats)),
                name=self.name,
                symmetries=self.symmetries,
                iterative=self.iterative,
            )
            .add_initial(strat.AddAssumptionFactory(), apply_first=True)
            .add_initial(strat.RearrangeAssumptionFactory(), apply_first=True)
        )

    def make_fusion(
        self,
        component: bool = False,
        tracked: bool = True,
        apply_first: bool = False,
        isolation_level: Optional[str] = None,
    ) -> "TileScopePack":
        """
        Create a new pack by adding fusion to the current pack.

        If component, it will add component fusion.
        If tracked, it will return the pack for finding a tracked tree.
        If apply_first, it will add fusion to the front of the initial strategies.
        """
        pack = self
        if component:
            pack = pack.add_initial(
                strat.ComponentFusionFactory(
                    tracked=tracked, isolation_level=isolation_level
                ),
                "{}_component_fusion{}".format(
                    "tracked" if tracked else "untracked",
                    "" if isolation_level is None else "_" + isolation_level,
                ),
                apply_first=apply_first,
            )
        else:
            pack = pack.add_initial(
                strat.FusionFactory(tracked=tracked, isolation_level=isolation_level),
                "{}_fusion{}".format(
                    "tracked" if tracked else "untracked",
                    "" if isolation_level is None else "_" + isolation_level,
                ),
                apply_first=apply_first,
            )
        if tracked:
            pack = pack.add_initial(strat.AddAssumptionFactory(), apply_first=True)
            if component:
                pack = pack.add_initial(
                    strat.DetectComponentsStrategy(ignore_parent=True), apply_first=True
                )
            pack = pack.add_initial(
                strat.RearrangeAssumptionFactory(), apply_first=True
            )
        return pack

    def make_interleaving(
        self, tracked: bool = True, unions: bool = False
    ) -> "TileScopePack":
        """
        Return a new pack where the factor strategy is replaced with an
        interleaving factor strategy.

        If unions is set to True it will overwrite unions on the strategy, and
        also pass the argument to AddInterleavingAssumption method.
        """

        def replace_list(strats):
            """Return a new list with the replaced tracked factor strategy."""
            res = []
            for strategy in strats:
                if isinstance(strategy, strat.FactorFactory):
                    d = strategy.to_jsonable()
                    d["interleaving"] = "all"
                    d["tracked"] = tracked
                    d["unions"] = d["unions"] or unions
                    res.append(AbstractStrategy.from_dict(d))
                else:
                    res.append(strategy)
            return res

        pack = self.__class__(
            ver_strats=replace_list(self.ver_strats),
            inferral_strats=replace_list(self.inferral_strats),
            initial_strats=replace_list(self.initial_strats),
            expansion_strats=list(map(replace_list, self.expansion_strats)),
            name=(
                self.name
                + "_{}_interleaving{}".format(
                    "tracked" if tracked else "untracked",
                    "_unions" if unions else "",
                )
            ),
            symmetries=self.symmetries,
            iterative=self.iterative,
        )

        if tracked:
            pack = pack.add_initial(
                strat.AddInterleavingAssumptionFactory(unions=unions), apply_first=True
            )
            pack = pack.add_initial(strat.AddAssumptionFactory(), apply_first=True)

        return pack

    def make_elementary(self) -> "TileScopePack":
        """
        Create a new pack by using only one by one and elementary
        verification.
        """
        if (
            strat.ElementaryVerificationStrategy() in self
            and strat.OneByOneVerificationStrategy() in self
            and len(self.ver_strats) == 2
            and self.iterative
        ):
            raise ValueError("The pack is already elementary.")
        pack = self.make_iterative()
        pack = pack.add_verification(strat.OneByOneVerificationStrategy(), replace=True)
        return pack.add_verification(
            strat.ElementaryVerificationStrategy(), "elementary"
        )

    def make_database(self) -> "TileScopePack":
        """
        Create a new pack by adding database verification to the current pack.
        """
        return self.add_verification(strat.DatabaseVerificationStrategy(), "database")

    def add_all_symmetry(self) -> "TileScopePack":
        """Create a new pack by turning on symmetry on the current pack."""
        if self.symmetries:
            raise ValueError("Symmetries already turned on.")
        return super().add_symmetry(strat.SymmetriesFactory(), "symmetries")

    # Creation of the base pack
    @classmethod
    def all_the_strategies(cls, length: int = 1) -> "TileScopePack":
        initial_strats: List[CSSstrategy] = [strat.FactorFactory()]
        if length > 1:
            initial_strats.append(strat.RequirementCorroborationFactory())

        return TileScopePack(
            initial_strats=initial_strats,
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [
                    strat.RequirementInsertionFactory(maxreqlen=length),
                    strat.AllPlacementsFactory(),
                ],
            ],
            name=f"all_the_strategies_{length}",
        )

    @classmethod
    def pattern_placements(
        cls, length: int = 1, partial: bool = False
    ) -> "TileScopePack":
        name = "{}{}{}_placements".format(
            "length_{}_".format(length) if length > 1 else "",
            "partial_" if partial else "",
            "pattern" if length > 1 else "pattern_point",
        )

        expansion_strats: List[CSSstrategy] = [
            strat.FactorFactory(unions=True),
            strat.CellInsertionFactory(maxreqlen=length),
        ]
        if length > 1:
            expansion_strats.append(strat.RequirementCorroborationFactory())

        return TileScopePack(
            initial_strats=[strat.PatternPlacementFactory(partial=partial)],
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[expansion_strats],
            name=name,
        )

    @classmethod
    def point_placements(
        cls, length: int = 1, partial: bool = False
    ) -> "TileScopePack":
        name = "{}{}point_placements".format(
            "length_{}_".format(length) if length > 1 else "",
            "partial_" if partial else "",
        )

        initial_strats: List[CSSstrategy] = [strat.FactorFactory()]
        if length > 1:
            initial_strats.append(strat.RequirementCorroborationFactory())

        return TileScopePack(
            initial_strats=initial_strats,
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [
                    strat.CellInsertionFactory(maxreqlen=length),
                    strat.PatternPlacementFactory(partial=partial),
                ],
            ],
            name=name,
        )

    @classmethod
    def insertion_point_placements(cls, partial: bool = False) -> "TileScopePack":
        name = "insertion_"
        partial_str = "partial_" if partial else ""
        name += f"{partial_str}point_placements"
        return TileScopePack(
            initial_strats=[
                strat.FactorFactory(),
                strat.CellInsertionFactory(
                    maxreqlen=1, ignore_parent=True, one_cell_only=True
                ),
            ],
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[[strat.PatternPlacementFactory(partial=partial)]],
            name=name,
        )

    @classmethod
    def regular_insertion_encoding(cls, direction: int) -> "TileScopePack":
        """This pack finds insertion encodings."""
        if direction not in DIRS:
            raise ValueError("Must be direction in {}.".format(DIRS))
        place_row = direction in (DIR_NORTH, DIR_SOUTH)
        place_col = not place_row
        name = "regular_insertion_encoding_{}".format(
            {
                DIR_EAST: "left",
                DIR_WEST: "right",
                DIR_NORTH: "bottom",
                DIR_SOUTH: "top",
            }[direction]
        )
        return TileScopePack(
            initial_strats=[
                strat.FactorFactory(),
                strat.CellInsertionFactory(ignore_parent=True, one_cell_only=True),
            ],
            ver_strats=[strat.BasicVerificationStrategy()],
            inferral_strats=[],
            expansion_strats=[
                [
                    strat.RowAndColumnPlacementFactory(
                        place_col=place_col, place_row=place_row, dirs=[direction]
                    )
                ]
            ],
            name=name,
        )

    @classmethod
    def row_and_col_placements(
        cls, row_only: bool = False, col_only: bool = False, partial: bool = False
    ) -> "TileScopePack":
        if row_only and col_only:
            raise ValueError("Can't be row and col only.")
        place_row = not col_only
        place_col = not row_only
        both = place_col and place_row
        name = "{}{}{}{}_placements".format(
            "partial_" if partial else "",
            "row" if not col_only else "",
            "_and_" if both else "",
            "col" if not row_only else "",
        )
        rowcol_strat = strat.RowAndColumnPlacementFactory(
            place_row=place_row, place_col=place_col, partial=partial
        )
        expansion_strats: List[List[CSSstrategy]] = [[rowcol_strat]]
        if partial:
            expansion_strats.append([strat.PatternPlacementFactory(point_only=True)])
        return TileScopePack(
            initial_strats=[strat.FactorFactory()],
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=expansion_strats,
            name=name,
        )

    @classmethod
    def insertion_row_and_col_placements(
        cls, row_only: bool = False, col_only: bool = False, partial: bool = False
    ) -> "TileScopePack":
        pack = cls.row_and_col_placements(row_only, col_only, partial)
        pack.name = "insertion_" + pack.name
        pack = pack.add_initial(
            strat.CellInsertionFactory(
                maxreqlen=1, ignore_parent=True, one_cell_only=True
            )
        )
        return pack

    @classmethod
    def insertion_point_row_and_col_placements(
        cls, row_only: bool = False, col_only: bool = False, partial: bool = False
    ) -> "TileScopePack":
        pack = cls.insertion_row_and_col_placements(row_only, col_only, partial)
        if not partial:
            # if partial, then already added pattern placements
            pack.name = pack.name.replace("insertion", "insertion_point")
            pack.expansion_strats = pack.expansion_strats + (
                (strat.PatternPlacementFactory(partial=False),),
            )
        return pack

    @classmethod
    def only_root_placements(
        cls,
        length: int = 3,
        max_num_req: Optional[int] = 1,
        max_placement_rules_per_req: Optional[int] = None,
        partial: bool = False,
    ) -> "TileScopePack":
        partial_str = "partial_" if partial else ""
        if max_num_req is not None:
            name = (
                f"only_length_{length}_{max_num_req}_reqs_root_{partial_str}placements"
            )
        else:
            name = f"only_length_{length}_root_{partial_str}placements"
        placement_factory = strat.RequirementPlacementFactory(
            max_rules_per_req=max_placement_rules_per_req, partial=partial
        )
        return TileScopePack(
            initial_strats=[
                strat.RootInsertionFactory(maxreqlen=length, max_num_req=max_num_req),
                strat.FactorFactory(unions=True, ignore_parent=False, workable=False),
            ],
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[[placement_factory]],
            name=name,
        )

    @classmethod
    def requirement_placements(
        cls, length: int = 2, partial: bool = False
    ) -> "TileScopePack":
        name = "{}{}requirement_placements".format(
            "length_{}_".format(length) if length != 2 else "",
            "partial_" if partial else "",
        )

        initial_strats: List[CSSstrategy] = [strat.FactorFactory()]
        if length > 1:
            initial_strats.append(strat.RequirementCorroborationFactory())

        return TileScopePack(
            initial_strats=initial_strats,
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [
                    strat.RequirementInsertionFactory(maxreqlen=length),
                    strat.PatternPlacementFactory(partial=partial),
                ],
            ],
            name=name,
        )

    @classmethod
    def point_and_row_and_col_placements(
        cls,
        length: int = 1,
        row_only: bool = False,
        col_only: bool = False,
        partial: bool = False,
    ) -> "TileScopePack":
        if row_only and col_only:
            raise ValueError("Can't be row and col only.")
        place_row = not col_only
        place_col = not row_only
        both = place_col and place_row
        name = "{}{}point_and_{}{}{}_placements".format(
            "length_{}_".format(length) if length > 1 else "",
            "partial_" if partial else "",
            "row" if not col_only else "",
            "_and_" if both else "",
            "col" if not row_only else "",
        )
        rowcol_strat = strat.RowAndColumnPlacementFactory(
            place_row=place_row, place_col=place_col, partial=partial
        )

        initial_strats: List[CSSstrategy] = [strat.FactorFactory()]
        if length > 1:
            initial_strats.append(strat.RequirementCorroborationFactory())

        return TileScopePack(
            initial_strats=initial_strats,
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [
                    strat.CellInsertionFactory(maxreqlen=length),
                    strat.PatternPlacementFactory(partial=partial),
                    rowcol_strat,
                ],
            ],
            name=name,
        )



import argparse
import inspect
from functools import partial
from typing import Callable, Dict

from logzero import logger

from permuta import Perm
from permuta.misc import DIR_SOUTH, DIR_WEST
from permuta.permutils import (
    is_insertion_encodable_maximum,
    is_insertion_encodable_rightmost,
)
from tilings import Tiling
from tilings.strategy_pack import TileScopePack
from tilings.tilescope import TileScope

PackBuilder = Callable[..., TileScopePack]

BASE_PACK: Dict[str, PackBuilder] = {
    "all_the_strategies": TileScopePack.all_the_strategies,
    "pattern_placements": TileScopePack.pattern_placements,
    "point_placements": TileScopePack.point_placements,
    "insertion_point_placements": TileScopePack.insertion_point_placements,
    "row_and_col_placements": TileScopePack.row_and_col_placements,
    "col_placements": partial(TileScopePack.row_and_col_placements, col_only=True),
    "row_placements": partial(TileScopePack.row_and_col_placements, row_only=True),
    "insertion_row_and_col_placements": TileScopePack.insertion_row_and_col_placements,
    "insertion_row_placements": partial(
        TileScopePack.insertion_row_and_col_placements, row_only=True
    ),
    "insertion_col_placements": partial(
        TileScopePack.insertion_row_and_col_placements, col_only=True
    ),
    "only_root_placements": TileScopePack.only_root_placements,
    "regular_insertion_encoding": TileScopePack.regular_insertion_encoding,
    "requirement_placements": TileScopePack.requirement_placements,
}


def list_stratpacks(args: argparse.Namespace) -> int:
    """
    Prints out every strategy pack available.
    """
    for pack in BASE_PACK:
        print(pack)
    return 0


def valid_kwarg_or_error(func, kwarg_name, pack_name):
    if kwarg_name not in inspect.signature(func).parameters:
        parser.error("Invalid argument {} for {}".format(kwarg_name, pack_name))


def build_pack(args: argparse.Namespace) -> TileScopePack:
    if args.strategy_pack not in BASE_PACK:
        parser.error(
            "Invalid strategy pack. Use 'tilescope list' to see available packs. "
            "Perhaps you got the order wrong? A valid command should be of "
            "the form 'tilescope spec {basis} {pack}'."
        )

    pack_builder: PackBuilder = BASE_PACK[args.strategy_pack]
    kwargs = dict()
    if args.length is not None:
        valid_kwarg_or_error(pack_builder, "length", args.strategy_pack)
        kwargs["length"] = args.length
    if args.strategy_pack == "regular_insertion_encoding":
        basis = [Perm.to_standard(p) for p in args.basis.split("_")]
        if is_insertion_encodable_maximum(basis):
            kwargs["direction"] = DIR_SOUTH
        elif is_insertion_encodable_rightmost(basis):
            kwargs["direction"] = DIR_WEST
        else:
            parser.error(
                "The basis does not have regular insertion encoding, "
                "so try another pack. Note: the extra args of the 'tilescope' "
                "command don't work with this pack, use instead the packs "
                "insertion_row_and_col_placements which expands in a similar "
                "fashion, but has more powerful strategies."
            )
    pack = pack_builder(**kwargs)
    if args.fusion:
        pack = pack.make_fusion()
    if args.database:
        pack = pack.make_database()
    if args.symmetries:
        pack = pack.add_all_symmetry()
    if args.elementary:
        pack = pack.make_elementary()
    return pack


def search_spec(args):
    """
    Search for a specification.
    """
    pack = build_pack(args)
    start_class = Tiling.from_string(args.basis)
    css = TileScope(start_class, pack)
    spec = css.auto_search(status_update=30)
    logger.info("The generating function is %s", spec.get_genf())
    return 0


parser = argparse.ArgumentParser(
    description="A command line tool for the TileScope algorithm."
)
subparsers = parser.add_subparsers(title="subcommands")

# List command
helpstr = "List all the strategy packs available"
parser_list = subparsers.add_parser("list", help=helpstr, description=helpstr)
parser_list.set_defaults(func=list_stratpacks)

# Spec command
helpstr = (
    "Search for a specification for a given permutation class with a given "
    "strategy pack."
)
parser_tree = subparsers.add_parser("spec", help=helpstr, description=helpstr)
parser_tree.add_argument(
    "basis",
    type=str,
    help="The basis of the permutation class. This can be 1- or 0-based and patterns"
    " should be separated by an underscore, e.g. 012_021.",
)
parser_tree.add_argument(
    "strategy_pack",
    type=str,
    help="The strategy pack to run. The strategy pack defines the set of "
    "strategies that will be used when searching for a specification. The "
    "command 'tilescope list' will show you the available packs.",
)
parser_tree.add_argument(
    "-l", "--length", type=int, help="Change the length parameter of the pack."
)
parser_tree.add_argument(
    "-f", "--fusion", action="store_true", help="Adds fusion to the pack."
)
parser_tree.add_argument(
    "-d",
    "--database",
    action="store_true",
    help="Adds database verification to the pack.",
)
parser_tree.add_argument(
    "-s", "--symmetries", action="store_true", help="Adds symmetries to the pack"
)
parser_tree.add_argument(
    "-e", "--elementary", action="store_true", help="Makes the pack elementary."
)
parser_tree.set_defaults(func=search_spec)


def main():
    args = parser.parse_args()
    if not hasattr(args, "func"):
        parser.error("Invalid command")
    return args.func(args)


if __name__ == "__main__":
    main()



from .enumeration import DatabaseEnumeration, LocalEnumeration, MonotoneTreeEnumeration
from .factor import Factor, FactorWithInterleaving, FactorWithMonotoneInterleaving
from .fusion import ComponentFusion, Fusion
from .gridded_perm_generation import GriddedPermsOnTiling
from .gridded_perm_reduction import GriddedPermReduction
from .guess_obstructions import guess_obstructions
from .minimal_gridded_perms import MinimalGriddedPerms
from .obstruction_inferral import (
    AllObstructionInferral,
    EmptyCellInferral,
    SubobstructionInferral,
)
from .obstruction_transitivity import ObstructionTransitivity
from .requirement_placement import RequirementPlacement
from .row_col_separation import RowColSeparation
from .sliding import Sliding
from .subclass_verification import SubclassVerificationAlgorithm

__all__ = [
    "DatabaseEnumeration",
    "LocalEnumeration",
    "MonotoneTreeEnumeration",
    "Factor",
    "FactorWithInterleaving",
    "FactorWithMonotoneInterleaving",
    "ComponentFusion",
    "Fusion",
    "MinimalGriddedPerms",
    "AllObstructionInferral",
    "EmptyCellInferral",
    "SubobstructionInferral",
    "ObstructionTransitivity",
    "GriddedPermsOnTiling",
    "GriddedPermReduction",
    "RequirementPlacement",
    "RowColSeparation",
    "SubclassVerificationAlgorithm",
    "guess_obstructions",
    "Sliding",
]



from typing import Callable, Iterator, List, Optional, Tuple

import sympy

from comb_spec_searcher import Constructor
from comb_spec_searcher.typing import (
    Parameters,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from tilings import GriddedPerm


class DummyConstructor(Constructor):
    def __init__(
        self,
    ):
        pass

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        raise NotImplementedError

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        raise NotImplementedError

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]], ...]]]:
        raise NotImplementedError

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ):
        raise NotImplementedError

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        raise NotImplementedError("Required for bijections")



from collections import defaultdict
from functools import partial
from itertools import chain, islice
from typing import Callable, Dict, Iterable, List, Optional, Set, Tuple

from ..griddedperm import GriddedPerm

Cell = Tuple[int, int]
Requirement = Tuple[GriddedPerm, ...]


class GriddedPermReduction:
    def __init__(
        self,
        obstructions: Tuple[GriddedPerm, ...],
        requirements: Tuple[Tuple[GriddedPerm, ...], ...],
        sorted_input: bool = False,
        already_minimized_obs: bool = False,
    ):
        # Only using MGP for typing purposes.
        if sorted_input:
            self._obstructions = obstructions
            self._requirements = requirements
        else:
            self._obstructions = tuple(sorted(obstructions))
            self._requirements = tuple(sorted(tuple(sorted(r)) for r in requirements))

        self._minimize_griddedperms(already_minimized_obs=already_minimized_obs)

    @property
    def obstructions(self) -> Tuple[GriddedPerm, ...]:
        return self._obstructions

    @property
    def requirements(self) -> Tuple[Requirement, ...]:
        return self._requirements

    def _minimize_griddedperms(self, already_minimized_obs=False) -> None:
        """Minimizes the set of obstructions and the set of requirement lists.
        The set of obstructions are first reduced to a minimal set. The
        requirements that contain any obstructions are removed from their
        respective lists. If any requirement list is empty, then the tiling is
        empty.
        """

        def set_empty():
            self._obstructions = (GriddedPerm.empty_perm(),)
            self._requirements = tuple()

        while True:
            # Minimize the set of obstructions
            minimized_obs = self.minimal_obs(
                already_minimized_obs=already_minimized_obs
            )

            if minimized_obs and not minimized_obs[0]:
                set_empty()
                break

            self._obstructions = tuple(sorted(minimized_obs))

            # Minimize the set of requiriments
            minimized_requirements = self.minimal_reqs(minimized_obs)
            minimized_requirements = tuple(
                sorted(tuple(sorted(set(req))) for req in minimized_requirements)
            )

            if minimized_requirements and not minimized_requirements[0]:
                set_empty()
                break

            if self._requirements == minimized_requirements:
                break

            self._requirements = minimized_requirements

    # if there is a requirement for which every component contains the same factor of
    # obstruction, then that factor can be removed from obstruction
    # [subobstruction inferral]
    def _clean_isolated(self, obstruction: GriddedPerm) -> GriddedPerm:
        """Remove the isolated factors that are implied by requirements
        from all obstructions."""
        cells_to_remove: Set[Cell] = set()
        for factor in obstruction.factors():
            if self._griddedperm_implied_by_some_requirement(factor):
                cells_to_remove.update(factor.pos)
        if cells_to_remove:
            obstruction = obstruction.remove_cells(cells_to_remove)
        return obstruction

    def minimal_obs(self, already_minimized_obs=False) -> Tuple[GriddedPerm, ...]:
        min_perms = (
            self._obstructions
            if already_minimized_obs
            else GriddedPermReduction._minimize(self._obstructions)
        )
        changed = []
        unchanged = []
        for ob in min_perms:
            cleaned_perm = self._clean_isolated(ob)
            if cleaned_perm == ob:
                unchanged.append(cleaned_perm)
            else:
                changed.append(cleaned_perm)
        if not changed:
            return tuple(unchanged)

        return GriddedPermReduction._minimize(changed) + tuple(
            gp for gp in unchanged if gp.avoids(*changed)
        )

    def minimal_reqs(
        self, obstructions: Tuple[GriddedPerm, ...]
    ) -> Tuple[Requirement, ...]:
        algos: Tuple[Callable, ...] = (
            GriddedPermReduction.factored_reqs,
            self.cleaned_requirements,
            partial(self.remove_avoided, obstructions=obstructions),
        )
        minimized_requirements = self._requirements
        for algo in algos:
            minimized_requirements = algo(minimized_requirements)
            if any(not r for r in minimized_requirements):
                return (tuple(),)
        return tuple(minimized_requirements)

    # If all of the gp's in a requirment list contain the same factor, you can remove
    # that factor from all of them and add it as it's own size one requirement list
    # [size one requirement list inferral]
    @staticmethod
    def factored_reqs(requirements: Iterable[Requirement]) -> List[Requirement]:
        """
        Add factors of requirements as size one requirements, removing them
        from each of the gridded permutations in the requirement it is
        contained in.
        """
        res: List[Requirement] = list()
        for requirement in requirements:
            if not requirement:
                # If req is empty, then this requirement can't be satisfied, so
                # return it and mark obstructions and requirements as empty.
                return [tuple()]
            # If any gridded permutation in list is empty then you vacuously
            # contain this requirement
            if all(requirement):
                factors = GriddedPermReduction.factors(requirement)
                if not factors or (len(factors) == 1 and len(requirement) == 1):
                    # if there are no factors in the intersection, or it is just
                    # the same req as the first, we do nothing and add the original
                    res.append(requirement)
                    continue
                # add each of the factors as a single requirement, and then remove
                # these from each of the other requirements in the list
                remaining_cells = set(c for gp in requirement for c in gp.pos) - set(
                    c for gp in factors for c in gp.pos
                )
                for factor in factors:
                    res.append((factor,))
                res.append(
                    tuple(
                        gp.get_gridded_perm_in_cells(remaining_cells)
                        for gp in requirement
                    )
                )
        return res

    # if a gridded perm G in a requirment has a factor that appears in EVERY gridded
    # perm of some other requirement, then that factor can be removed from G
    # [subrequirement inferral]
    # NOTE: This handles the following case:
    #   requirement R2 makes requirement R1 redundant if:
    #   For all G in R2, there exists H in R1 such that H <= G
    def cleaned_requirements(
        self, requirements: List[Requirement]
    ) -> List[Requirement]:
        """
        Remove the factors of a gridded permutation in a requirement if it is
        implied by another requirement.
        """
        for idx, requirement in enumerate(requirements):
            if not all(requirement):
                continue
            newgps: List[GriddedPerm] = []
            for gp in requirement:
                cells: List[Cell] = []
                for f in gp.factors():
                    if not self._griddedperm_implied_by_some_requirement(
                        f,
                        chain(
                            islice(requirements, idx),
                            islice(requirements, idx + 1, None),
                        ),
                    ):
                        cells.extend(f.pos)
                newgp = gp.get_gridded_perm_in_cells(cells)
                newgps.append(newgp)
            requirements[idx] = tuple(newgps)
        return requirements

    # We can remove any gridded perm in any requirement if the gridded perm contains
    # an obstruction
    # [requirement component deletion]
    def remove_avoided(
        self,
        requirements: List[Requirement],
        obstructions: Optional[Tuple[GriddedPerm, ...]] = None,
    ) -> List[Requirement]:
        if obstructions is None:
            obstructions = self._obstructions
        res: List[Requirement] = []
        for requirement in requirements:
            # If any gridded permutation in list is empty then you vacuously
            # contain this requirement
            if not all(requirement):
                continue
            cleanreq = tuple(
                gp
                for gp in GriddedPermReduction._minimize(requirement)
                if gp.avoids(*obstructions)
            )
            # If cleanreq is empty, then can not contain this requirement so
            # the tiling is empty.
            if not cleanreq:
                return [tuple()]
            if not (len(cleanreq) == 1 and not cleanreq[0]):
                res.append(cleanreq)
        return res

    @staticmethod
    def _griddedperm_implied_by_requirement(
        griddedperm: GriddedPerm, requirement: Iterable[GriddedPerm]
    ):
        """
        Return True, if the containment of the gridded perm is implied by
        the containment of the requirement.
        """
        return all(griddedperm in gp for gp in requirement)

    # detect if there exists a requirement such that every component in that requirement
    # contains griddedperm
    def _griddedperm_implied_by_some_requirement(
        self,
        griddedperm: GriddedPerm,
        requirements: Optional[Iterable[Requirement]] = None,
    ):
        """
        Return True if the containiment some requirement implies the
        containment griddedperm.
        """
        if requirements is None:
            requirements = self._requirements
        return any(
            GriddedPermReduction._griddedperm_implied_by_requirement(
                griddedperm, requirement
            )
            for requirement in requirements
        )

    @staticmethod
    def _minimize(griddedperms: Iterable[GriddedPerm]) -> Tuple[GriddedPerm, ...]:
        """
        Removes non-minimal gridded permutations from the set.
        """
        perms_by_size: Dict[int, List[GriddedPerm]] = defaultdict(list)
        for gp in griddedperms:
            perms_by_size[len(gp)].append(gp)
        sizes = sorted(perms_by_size.keys())
        if not sizes:
            return tuple()
        minimal_perms = set(perms_by_size[sizes[0]])
        for size in sizes[1:]:
            next_layer = set()
            for gp in perms_by_size[size]:
                if gp.avoids(*minimal_perms):
                    next_layer.add(gp)
            minimal_perms |= next_layer
        return tuple(minimal_perms)

    @staticmethod
    def factors(griddedperms: Tuple[GriddedPerm, ...]) -> Set[GriddedPerm]:
        res: Set[GriddedPerm] = set(griddedperms[0].factors())
        for gp in griddedperms[1:]:
            if not res:
                break
            res = res.intersection(gp.factors())
        return res



from collections import Counter
from functools import reduce
from itertools import chain
from operator import mul
from typing import Callable, Dict, Iterable, Iterator, List, Optional, Set, Tuple, cast

from sympy import Eq, Function

from comb_spec_searcher import (
    CartesianProduct,
    CartesianProductStrategy,
    StrategyFactory,
)
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.typing import (
    Parameters,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from permuta import Perm
from tilings import GriddedPerm, Tiling
from tilings.algorithms import (
    Factor,
    FactorWithInterleaving,
    FactorWithMonotoneInterleaving,
)
from tilings.assumptions import TrackingAssumption
from tilings.exception import InvalidOperationError
from tilings.misc import multinomial, partitions_iterator

Cell = Tuple[int, int]

__all__ = (
    "FactorFactory",
    "FactorStrategy",
    "FactorWithInterleavingStrategy",
    "FactorWithMonotoneInterleavingStrategy",
)


class FactorStrategy(CartesianProductStrategy[Tiling, GriddedPerm]):
    def __init__(
        self,
        partition: Iterable[Iterable[Cell]],
        ignore_parent: bool = True,
        workable: bool = True,
    ):
        self.partition = tuple(sorted(tuple(sorted(p)) for p in partition))
        inferrable = any(interleaving_rows_and_cols(self.partition))
        super().__init__(
            ignore_parent=ignore_parent, workable=workable, inferrable=inferrable
        )

    def decomposition_function(self, tiling: Tiling) -> Tuple[Tiling, ...]:
        return tuple(tiling.sub_tiling(cells) for cells in self.partition)

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        extra_parameters: Tuple[Dict[str, str], ...] = tuple({} for _ in children)
        for parent_var, assumption in zip(
            comb_class.extra_parameters, comb_class.assumptions
        ):
            for idx, child in enumerate(children):
                # TODO: consider skew/sum
                new_assumption = child.forward_map_assumption(assumption)
                if new_assumption.gps:
                    child_var = child.get_assumption_parameter(new_assumption)
                    extra_parameters[idx][parent_var] = child_var
        return extra_parameters

    def formal_step(self) -> str:
        """
        Return a string that describe the operation performed on the tiling.
        """
        return "factor with partition {}".format(
            " / ".join(
                "{{{}}}".format(", ".join(str(c) for c in part))
                for part in self.partition
            )
        )

    def backward_map(
        self,
        tiling: Tiling,
        gps: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(tiling)
        gps_to_combine = tuple(
            tiling.backward_map(cast(GriddedPerm, gp))
            for gp, tiling in zip(gps, children)
        )
        temp = [
            ((cell[0], idx), (cell[1], val))
            for gp in gps_to_combine
            for (idx, val), cell in zip(enumerate(gp.patt), gp.pos)
        ]
        temp.sort()
        new_pos = [(idx[0], val[0]) for idx, val in temp]
        new_patt = Perm.to_standard(val for _, val in temp)
        yield GriddedPerm(new_patt, new_pos)

    def forward_map(
        self,
        tiling: Tiling,
        gp: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[GriddedPerm, ...]:
        if children is None:
            children = self.decomposition_function(tiling)
        return tuple(
            tiling.forward_map(gp.get_gridded_perm_in_cells(part))
            for tiling, part in zip(children, self.partition)
        )

    def __str__(self) -> str:
        return "factor"

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"partition={self.partition}",
                f"ignore_parent={self.ignore_parent}",
                f"workable={self.workable}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    # JSON methods

    def to_jsonable(self) -> dict:
        """Return a dictionary form of the strategy."""
        d: dict = super().to_jsonable()
        d.pop("inferrable")
        d.pop("possibly_empty")
        d["partition"] = self.partition
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "FactorStrategy":
        partition = cast(
            Tuple[Tuple[Cell]],
            tuple(tuple(tuple(c) for c in p) for p in d.pop("partition")),
        )
        return cls(partition=partition, **d)


# The following functions are used to determine assumptions needed to count the
# interleavings of a factor. They are also used by AddInterleavingAssumptionStrategy.


def interleaving_rows_and_cols(
    partition: Tuple[Tuple[Cell, ...], ...]
) -> Tuple[Set[int], Set[int]]:
    """
    Return the set of cols and the set of rows that are being interleaved when
    factoring with partition.
    """
    cols: Set[int] = set()
    rows: Set[int] = set()
    x_seen: Set[int] = set()
    y_seen: Set[int] = set()
    for part in partition:
        cols.update(x for x, _ in part if x in x_seen)
        rows.update(y for _, y in part if y in y_seen)
        x_seen.update(x for x, _ in part)
        y_seen.update(y for _, y in part)
    return cols, rows


def assumptions_to_add(
    cells: Tuple[Cell, ...], cols: Set[int], rows: Set[int]
) -> Tuple[TrackingAssumption, ...]:
    """
    Return the assumptions that should be tracked in the set of cells if we are
    interleaving the given rows and cols.
    """
    col_assumptions = [
        TrackingAssumption(
            [GriddedPerm.point_perm(cell) for cell in cells if x == cell[0]]
        )
        for x in cols
    ]
    row_assumptions = [
        TrackingAssumption(
            [GriddedPerm.point_perm(cell) for cell in cells if y == cell[1]]
        )
        for y in rows
    ]
    return tuple(ass for ass in chain(col_assumptions, row_assumptions) if ass.gps)


def contains_interleaving_assumptions(
    comb_class: Tiling, partition: Tuple[Tuple[Cell, ...], ...]
) -> bool:
    """
    Return True if the parent tiling contains all of the necessary tracking
    assumptions needed to count the interleavings, and therefore the
    children too.
    """
    cols, rows = interleaving_rows_and_cols(partition)
    return all(
        ass in comb_class.assumptions
        for ass in chain.from_iterable(
            assumptions_to_add(cells, cols, rows) for cells in partition
        )
    )


class Interleaving(CartesianProduct[Tiling, GriddedPerm]):
    def __init__(
        self,
        parent: Tiling,
        children: Iterable[Tiling],
        extra_parameters: Tuple[Dict[str, str], ...],
        interleaving_parameters: Iterable[Tuple[str, ...]],
    ):
        super().__init__(parent, children, extra_parameters)
        self.interleaving_parameters = tuple(interleaving_parameters)
        self.interleaving_indices = tuple(
            tuple(parent.extra_parameters.index(k) for k in parameters)
            for parameters in interleaving_parameters
        )

    @staticmethod
    def is_equivalence() -> bool:
        return False

    @staticmethod
    def get_equation(lhs_func: Function, rhs_funcs: Tuple[Function, ...]) -> Eq:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        non_interleaved_terms = super().get_terms(parent_terms, subterms, n)
        interleaved_terms: Terms = Counter()
        for parameters, value in non_interleaved_terms.items():
            # multinomial counts the number of ways to interleave the values k1, ...,kn.
            multiplier = reduce(
                mul,
                [
                    multinomial([parameters[k] for k in int_parameters])
                    for int_parameters in self.interleaving_indices
                ],
                1,
            )
            interleaved_terms[parameters] += multiplier * value
        return interleaved_terms

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]], ...]]]:
        raise NotImplementedError

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ):
        raise NotImplementedError


class FactorWithInterleavingStrategy(FactorStrategy):
    def formal_step(self) -> str:
        return "interleaving " + super().formal_step()

    def constructor(
        self, tiling: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Interleaving:
        if children is None:
            children = self.decomposition_function(tiling)
        try:
            interleaving_parameters = self.interleaving_parameters(tiling)
        except ValueError as e:
            # must be untracked
            raise NotImplementedError("The interleaving factor was not tracked.") from e
        return Interleaving(
            tiling,
            children,
            self.extra_parameters(tiling, children),
            interleaving_parameters,
        )

    def interleaving_parameters(self, comb_class: Tiling) -> List[Tuple[str, ...]]:
        """
        Return the parameters on the parent tiling that needed to be interleaved.
        """
        res: List[Tuple[str, ...]] = []
        cols, rows = interleaving_rows_and_cols(self.partition)
        for x in cols:
            assumptions = [
                TrackingAssumption(
                    GriddedPerm.point_perm(cell) for cell in cells if x == cell[0]
                )
                for cells in self.partition
            ]
            res.append(
                tuple(
                    comb_class.get_assumption_parameter(ass)
                    for ass in assumptions
                    if ass.gps
                )
            )
        for y in rows:
            assumptions = [
                TrackingAssumption(
                    GriddedPerm.point_perm(cell) for cell in cells if y == cell[1]
                )
                for cells in self.partition
            ]
            res.append(
                tuple(
                    comb_class.get_assumption_parameter(ass)
                    for ass in assumptions
                    if ass.gps
                )
            )
        return res

    def backward_map(
        self,
        tiling: Tiling,
        gps: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError

    def forward_map(
        self,
        tiling: Tiling,
        gp: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[GriddedPerm, ...]:
        raise NotImplementedError

    @staticmethod
    def get_eq_symbol() -> str:
        return "="

    @staticmethod
    def get_op_symbol() -> str:
        return "*"


class MonotoneInterleaving(Interleaving):
    pass


class FactorWithMonotoneInterleavingStrategy(FactorWithInterleavingStrategy):
    def constructor(
        self, tiling: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> MonotoneInterleaving:
        if children is None:
            children = self.decomposition_function(tiling)
        try:
            interleaving_parameters = self.interleaving_parameters(tiling)
        except ValueError as e:
            # must be untracked
            raise NotImplementedError(
                "The monotone interleaving was not tracked."
            ) from e
        return MonotoneInterleaving(
            tiling,
            children,
            self.extra_parameters(tiling, children),
            interleaving_parameters,
        )


class FactorFactory(StrategyFactory[Tiling]):

    FACTOR_ALGO_AND_CLASS = {
        None: (Factor, FactorStrategy),
        "monotone": (
            FactorWithMonotoneInterleaving,
            FactorWithMonotoneInterleavingStrategy,
        ),
        "all": (FactorWithInterleaving, FactorWithInterleavingStrategy),
    }

    def __init__(
        self,
        interleaving: Optional[str] = None,
        unions: bool = False,
        ignore_parent: bool = True,
        workable: bool = True,
        tracked: bool = False,
    ) -> None:
        try:
            self.factor_algo, self.factor_class = self.FACTOR_ALGO_AND_CLASS[
                interleaving
            ]
        except KeyError as e:
            raise InvalidOperationError(
                "Invalid interleaving option. Must be in "
                f"{FactorFactory.FACTOR_ALGO_AND_CLASS.keys()}, "
                f"used {interleaving}"
            ) from e
        self.unions = unions
        self.ignore_parent = ignore_parent
        self.workable = workable
        self.tracked = tracked and self.factor_class in (
            FactorWithInterleavingStrategy,
            FactorWithMonotoneInterleaving,
        )

    def __call__(self, comb_class: Tiling) -> Iterator[FactorStrategy]:
        factor_algo = self.factor_algo(comb_class)
        if factor_algo.factorable():
            min_comp = tuple(tuple(part) for part in factor_algo.get_components())
            if self.unions:
                for partition in partitions_iterator(min_comp):
                    components = tuple(
                        tuple(chain.from_iterable(part)) for part in partition
                    )
                    if not self.tracked or contains_interleaving_assumptions(
                        comb_class, components
                    ):
                        yield self._build_strategy(components, workable=False)
            if not self.tracked or contains_interleaving_assumptions(
                comb_class, min_comp
            ):
                yield self._build_strategy(min_comp, workable=self.workable)

    def _build_strategy(
        self, components: Tuple[Tuple[Cell, ...], ...], workable: bool
    ) -> FactorStrategy:
        """
        Build the factor strategy for the given components.

        It ensure that a plain factor rule is returned.
        """
        interleaving = any(interleaving_rows_and_cols(components))
        factor_strat = self.factor_class if interleaving else FactorStrategy
        return factor_strat(
            components, ignore_parent=self.ignore_parent, workable=workable
        )

    def __str__(self) -> str:
        if self.factor_class is FactorStrategy:
            s = "factor"
        elif self.factor_class is FactorWithInterleavingStrategy:
            s = "factor with interleaving"
        elif self.factor_class is FactorWithMonotoneInterleavingStrategy:
            s = "factor with monotone interleaving"
        else:
            raise Exception("Invalid interleaving type")
        if self.unions:
            s = "unions of " + s
        if self.tracked:
            s = "tracked " + s
        return s

    def __repr__(self) -> str:
        if self.factor_class is FactorStrategy:
            interleaving = None
        elif self.factor_class is FactorWithInterleavingStrategy:
            interleaving = "all"
        elif self.factor_class is FactorWithMonotoneInterleavingStrategy:
            interleaving = "monotone"
        args = ", ".join(
            [
                f"interleaving={interleaving!r}",
                f"unions={self.unions}",
                f"ignore_parent={self.ignore_parent}",
                f"workable={self.workable}",
                f"tracked={self.tracked}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        if self.factor_class is FactorStrategy:
            interleaving = None
        elif self.factor_class is FactorWithInterleavingStrategy:
            interleaving = "all"
        elif self.factor_class is FactorWithMonotoneInterleavingStrategy:
            interleaving = "monotone"
        d["interleaving"] = interleaving
        d["unions"] = self.unions
        d["ignore_parent"] = self.ignore_parent
        d["workable"] = self.workable
        d["tracked"] = self.tracked
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "FactorFactory":
        return cls(**d)



import abc
from itertools import chain, product
from typing import Dict, Iterable, Iterator, List, Optional, Tuple, cast

from comb_spec_searcher import DisjointUnionStrategy, StrategyFactory
from comb_spec_searcher.exception import StrategyDoesNotApply
from permuta import Av, Perm
from tilings import GriddedPerm, Tiling

ListRequirement = Tuple[GriddedPerm, ...]

EXTRA_BASIS_ERR = "'extra_basis' should be a list of Perm to avoid"

__all__ = [
    "CellInsertionFactory",
    "RootInsertionFactory",
    "RequirementExtensionFactory",
    "RequirementInsertionFactory",
    "FactorInsertionFactory",
    "RequirementCorroborationFactory",
]


class RequirementInsertionStrategy(DisjointUnionStrategy[Tiling, GriddedPerm]):
    def __init__(self, gps: Iterable[GriddedPerm], ignore_parent: bool = False):
        super().__init__(ignore_parent=ignore_parent)
        self.gps = frozenset(gps)

    def decomposition_function(self, tiling: Tiling) -> Tuple[Tiling, Tiling]:
        """
        Return a tuple of tiling. The first one avoids all the pattern in the
        list while the other contain one of the patterns in the list.
        """
        return tiling.add_obstructions(self.gps), tiling.add_list_requirement(self.gps)

    def formal_step(self) -> str:
        """
        Return the formal step for the insertion according to the req_list
        inserted.

        This needs to be redefined if you want to insert list requirement with
        more than one requirement.
        """
        if len(self.gps) == 1:
            req = tuple(self.gps)[0]
            if req.is_localized():
                return "insert {} in cell {}".format(req.patt, req.pos[0])
            return "insert {}".format(req)
        raise NotImplementedError

    def backward_map(
        self,
        tiling: Tiling,
        gps: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(tiling)
        idx = DisjointUnionStrategy.backward_map_index(gps)
        yield children[idx].backward_map(cast(GriddedPerm, gps[idx]))

    def forward_map(
        self,
        tiling: Tiling,
        gp: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], Optional[GriddedPerm]]:
        if children is None:
            children = self.decomposition_function(tiling)
        if gp.avoids(*self.gps):
            return (children[0].forward_map(gp), None)
        return (None, children[1].forward_map(gp))

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if not comb_class.extra_parameters:
            return super().extra_parameters(comb_class, children)
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        av, co = children
        av_params: Dict[str, str] = {}
        co_params: Dict[str, str] = {}
        for assumption in comb_class.assumptions:
            parent_var = comb_class.get_assumption_parameter(assumption)
            av_mapped_assumption = av.forward_map_assumption(assumption)
            if av_mapped_assumption.gps:
                child_var = av.get_assumption_parameter(av_mapped_assumption)
                av_params[parent_var] = child_var
            co_mapped_assumption = co.forward_map_assumption(assumption)
            if co_mapped_assumption.gps:
                child_var = co.get_assumption_parameter(co_mapped_assumption)
                co_params[parent_var] = child_var
        return av_params, co_params

    def __repr__(self) -> str:
        return "RequirementInsertionStrategy(gps={}, ignore_parent={})".format(
            self.gps, self.ignore_parent
        )

    def __str__(self) -> str:
        return "requirement insertion"

    def to_jsonable(self) -> dict:
        """Return a dictionary form of the strategy."""
        d: dict = super().to_jsonable()
        d.pop("inferrable")
        d.pop("possibly_empty")
        d.pop("workable")
        d["gps"] = [gp.to_jsonable() for gp in self.gps]
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RequirementInsertionStrategy":
        gps = [GriddedPerm.from_dict(gp) for gp in d.pop("gps")]
        return cls(gps=gps, **d)


class AbstractRequirementInsertionFactory(StrategyFactory[Tiling]):
    """
    Bases class for requirement insertion on tilings.

    It will create batch rules based on the containment or not of
    requirements.  The requirement  used are  provided by `req_list_to_insert`
    """

    def __init__(self, ignore_parent: bool = False):
        self.ignore_parent = ignore_parent

    @abc.abstractmethod
    def req_lists_to_insert(self, tiling: Tiling) -> Iterator[ListRequirement]:
        """
        Iterator over all the requirement list to insert to create the batch
        rules.
        """

    def __call__(self, comb_class: Tiling) -> Iterator[RequirementInsertionStrategy]:
        """
        Iterator over all the requirement insertion rules.
        """
        for req_list in self.req_lists_to_insert(comb_class):
            yield RequirementInsertionStrategy(req_list, self.ignore_parent)

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["ignore_parent"] = self.ignore_parent
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AbstractRequirementInsertionFactory":
        return cls(**d)

    def __repr__(self) -> str:
        return "{}(ignore_parent={})".format(
            self.__class__.__name__, self.ignore_parent
        )


class RequirementInsertionWithRestrictionFactory(AbstractRequirementInsertionFactory):
    """
    As RequirementInsertion, but a set of pattern to avoids and a maximum
    length can be provided.
    """

    def __init__(
        self,
        maxreqlen: int,
        extra_basis: Optional[List[Perm]] = None,
        ignore_parent: bool = False,
    ):
        if extra_basis is None:
            self.extra_basis = []
        else:
            assert isinstance(extra_basis, list), EXTRA_BASIS_ERR
            assert all(isinstance(p, Perm) for p in extra_basis), EXTRA_BASIS_ERR
            self.extra_basis = extra_basis
        self.maxreqlen = maxreqlen
        super().__init__(ignore_parent)

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["maxreqlen"] = self.maxreqlen
        d["extra_basis"] = self.extra_basis
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RequirementInsertionWithRestrictionFactory":
        if d["extra_basis"] is None:
            extra_basis = None
        else:
            extra_basis = [Perm(p) for p in d["extra_basis"]]
        d.pop("extra_basis")
        return cls(extra_basis=extra_basis, **d)

    def __repr__(self) -> str:
        return "{}(maxreqlen={}, extra_basis={}, " "ignore_parent={})".format(
            self.__class__.__name__,
            self.maxreqlen,
            self.extra_basis,
            self.ignore_parent,
        )


class CellInsertionFactory(RequirementInsertionWithRestrictionFactory):
    """
    The cell insertion strategy.

    The cell insertion strategy is a disjoint union strategy.
    For each active cell, the strategy considers all patterns (up to some maximum
    length given by `maxreqlen`) and returns two tilings; one which requires the
    pattern in the cell and one where the pattern is obstructed.

    The one_cell_only flag will ensure that the strategy only inserts into the
    'smallest' non-positive cell. This is used for 'insertion' packs where
    we are intending to make every cell positive, so with this setting we have
    a unique path to the fully positive tilings.
    """

    def __init__(
        self,
        maxreqlen: int = 1,
        extra_basis: Optional[List[Perm]] = None,
        ignore_parent: bool = False,
        one_cell_only: bool = False,
    ) -> None:
        super().__init__(maxreqlen, extra_basis, ignore_parent)
        self.one_cell_only = one_cell_only

    def req_lists_to_insert(self, tiling: Tiling) -> Iterator[ListRequirement]:
        if self.one_cell_only:
            assert self.maxreqlen == 1 and self.ignore_parent
            cells = sorted(
                frozenset(tiling.active_cells) - frozenset(tiling.positive_cells)
            )
            if cells:
                yield (GriddedPerm.single_cell((0,), cells[0]),)
            return

        active = tiling.active_cells
        bdict = tiling.cell_basis()
        for cell, length in product(active, range(1, self.maxreqlen + 1)):
            basis = bdict[cell][0] + self.extra_basis
            patterns = Av(basis).of_length(length) if basis else Perm.of_length(length)
            yield from (
                (GriddedPerm.single_cell(patt, cell),)
                for patt in patterns
                if not any(patt in perm for perm in bdict[cell][1])
            )

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["one_cell_only"] = self.one_cell_only
        return d

    def __str__(self) -> str:
        if self.maxreqlen == 1:
            return "point insertion"
        if self.extra_basis:
            perm_class = Av(self.extra_basis)
            return "cell insertion from {} up to " "length {}".format(
                perm_class, self.maxreqlen
            )
        return "cell insertion up to length {}".format(self.maxreqlen)

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"maxreqlen={self.maxreqlen}",
                f"extra_basis={self.extra_basis!r}",
                f"ignore_parent={self.ignore_parent}",
                f"one_cell_only={self.one_cell_only}",
            ]
        )
        return f"{self.__class__.__name__}({args})"


class RootInsertionFactory(CellInsertionFactory):
    """
    The cell insertion strategy performed only on 1 by 1 tilings.
    """

    def __init__(
        self,
        maxreqlen: int = 3,
        extra_basis: Optional[List[Perm]] = None,
        ignore_parent: bool = False,
        max_num_req: Optional[int] = None,
    ) -> None:
        super().__init__(maxreqlen, extra_basis, ignore_parent)
        self.max_num_req = max_num_req

    def __call__(self, comb_class: Tiling) -> Iterator[RequirementInsertionStrategy]:
        if comb_class.dimensions != (1, 1):
            return
        for strategy in super().__call__(comb_class):
            t_with_new_req = comb_class.add_list_requirement(strategy.gps)
            if (
                self.max_num_req is None
                or len(t_with_new_req.requirements) <= self.max_num_req
            ):
                yield strategy

    def __str__(self) -> str:
        if not self.extra_basis:
            s = "root insertion up to length {}".format(self.maxreqlen)
        else:
            perm_class = Av(self.extra_basis)
            s = "root insertion from {} up to length {}".format(
                perm_class, self.maxreqlen
            )
        if self.max_num_req is not None:
            s += " (up to {} requirements)".format(self.max_num_req)
        return s

    def __repr__(self) -> str:
        return (
            "{}(maxreqlen={}, extra_basis={}, "
            "ignore_parent={}, max_num_req={})".format(
                self.__class__.__name__,
                self.maxreqlen,
                self.extra_basis,
                self.ignore_parent,
                self.max_num_req,
            )
        )

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["max_num_req"] = self.max_num_req
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RootInsertionFactory":
        if d["extra_basis"] is None:
            extra_basis = None
        else:
            extra_basis = [Perm(p) for p in d["extra_basis"]]
        d.pop("extra_basis")
        d.pop("one_cell_only")
        return cls(extra_basis=extra_basis, **d)


class RequirementExtensionFactory(RequirementInsertionWithRestrictionFactory):
    """
    Insert longer requirements in to cells which contain a requirement
    """

    def __init__(
        self,
        maxreqlen: int = 2,
        extra_basis: Optional[List[Perm]] = None,
        ignore_parent: bool = False,
    ) -> None:
        super().__init__(maxreqlen, extra_basis, ignore_parent)

    def req_lists_to_insert(self, tiling: Tiling) -> Iterator[ListRequirement]:
        bdict = tiling.cell_basis()
        cell_with_req = (
            (cell, obs, reqlist[0])
            for cell, (obs, reqlist) in bdict.items()
            if len(reqlist) == 1
        )
        for cell, obs, curr_req in cell_with_req:
            for length in range(len(curr_req) + 1, self.maxreqlen + 1):
                basis = obs + self.extra_basis
                patterns = (
                    Av(basis).of_length(length) if basis else Perm.of_length(length)
                )
                for patt in patterns:
                    if curr_req in patt:
                        yield (GriddedPerm.single_cell(patt, cell),)

    def __str__(self) -> str:
        if self.extra_basis:
            perm_class = Av(self.extra_basis)
            return "requirement extension from {} up to " "length {}".format(
                perm_class, self.maxreqlen
            )
        return "requirement extension insertion up to " "length {}".format(
            self.maxreqlen
        )


class RequirementInsertionFactory(RequirementInsertionWithRestrictionFactory):
    """
    Insert all possible requirements the obstruction allows if the tiling does
    not have requirements.

    If <limited_insertion> is true, the default behavior, requirements will only be
    inserted on Tilings that have no requirements.
    """

    def __init__(
        self,
        maxreqlen: int = 2,
        extra_basis: Optional[List[Perm]] = None,
        limited_insertion: bool = True,
        ignore_parent: bool = False,
    ) -> None:
        self.limited_insertion = limited_insertion
        super().__init__(maxreqlen, extra_basis, ignore_parent)

    def req_lists_to_insert(self, tiling: Tiling) -> Iterator[ListRequirement]:
        obs_tiling = Tiling(
            tiling.obstructions,
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
            sorted_input=True,
        )
        for length in range(1, self.maxreqlen + 1):
            for gp in obs_tiling.gridded_perms_of_length(length):
                if len(gp.factors()) == 1 and all(
                    p not in gp.patt for p in self.extra_basis
                ):
                    yield (GriddedPerm(gp.patt, gp.pos),)

    def __call__(self, comb_class: Tiling) -> Iterator[RequirementInsertionStrategy]:
        if self.limited_insertion and comb_class.requirements:
            return
        yield from super().__call__(comb_class)

    def __str__(self) -> str:
        if self.maxreqlen == 1:
            return "point insertion"
        if self.extra_basis:
            perm_class = Av(self.extra_basis)
            return "requirement insertion from {} up to " "length {}".format(
                perm_class, self.maxreqlen
            )
        return "requirement insertion up to " "length {}".format(self.maxreqlen)

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"maxreqlen={self.maxreqlen}",
                f"extra_basis={self.extra_basis!r}",
                f"limited_insertion={self.limited_insertion}",
                f"ignore_parent={self.ignore_parent}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["limited_insertion"] = self.limited_insertion
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RequirementInsertionWithRestrictionFactory":
        if d["limited_insertion"] is None:
            extra_basis = None
        else:
            extra_basis = [Perm(p) for p in d["extra_basis"]]
        d.pop("extra_basis")
        limited_insertion = d.pop("limited_insertion")
        maxreqlen = d.pop("maxreqlen")
        return cls(
            maxreqlen=maxreqlen,
            extra_basis=extra_basis,
            limited_insertion=limited_insertion,
            **d,
        )


class FactorInsertionFactory(AbstractRequirementInsertionFactory):
    """
    Insert all proper factor of the requirement or obstructions on the tiling.
    """

    def __init__(self, ignore_parent: bool = True) -> None:
        super().__init__(ignore_parent)

    def req_lists_to_insert(self, tiling: Tiling) -> Iterator[ListRequirement]:
        reqs_and_obs: Iterator[GriddedPerm] = chain(
            tiling.obstructions, *tiling.requirements
        )
        gp_facts = map(GriddedPerm.factors, reqs_and_obs)
        proper_facts = chain.from_iterable(f for f in gp_facts if len(f) > 1)
        for f in proper_facts:
            yield (GriddedPerm(f.patt, f.pos),)

    def __str__(self) -> str:
        return "all factor insertion"


class RequirementCorroborationFactory(AbstractRequirementInsertionFactory):
    """
    The requirement corroboration strategy.

    The requirement corroboration strategy is a batch strategy that considers
    each requirement of each requirement list. For each of these requirements,
    the strategy returns two tilings; one where the requirement has been turned
    into an obstruction and another where the requirement has been singled out
    and a new requirement list added with only the requirement. This new
    requirement list contains only the singled out requirement.

    This implements the notion of partitioning the set of gridded permutations
    into those that satisfy this requirement and those that avoid it. Those
    that avoid the requirement, must therefore satisfy another requirement from
    the same list and hence the requirement list must be of length at least 2.
    """

    def __init__(self, ignore_parent: bool = True):
        super().__init__(ignore_parent)

    def req_lists_to_insert(self, tiling: Tiling) -> Iterator[ListRequirement]:
        to_insert = chain.from_iterable(
            reqs for reqs in tiling.requirements if len(reqs) > 1
        )
        for req in to_insert:
            yield (req,)

    def __str__(self) -> str:
        return "requirement corroboration"



import abc
from collections import defaultdict
from functools import reduce
from itertools import chain, product
from typing import Dict, Iterable, Iterator, List, Optional, Set, Tuple, cast

from comb_spec_searcher import DisjointUnionStrategy, StrategyFactory
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies import Rule
from permuta.misc import DIR_EAST, DIR_NORTH, DIR_SOUTH, DIR_WEST, DIRS
from tilings import GriddedPerm, Tiling
from tilings.algorithms import RequirementPlacement

__all__ = [
    "PatternPlacementFactory",
    "RequirementPlacementFactory",
    "RowAndColumnPlacementFactory",
    "AllPlacementsFactory",
]

Cell = Tuple[int, int]

DIR_REPR = {
    DIR_NORTH: "DIR_NORTH",
    DIR_SOUTH: "DIR_SOUTH",
    DIR_WEST: "DIR_WEST",
    DIR_EAST: "DIR_EAST",
}

DIR_STR = {
    DIR_NORTH: "north",
    DIR_SOUTH: "south",
    DIR_EAST: "east",
    DIR_WEST: "west",
}


class RequirementPlacementStrategy(DisjointUnionStrategy[Tiling, GriddedPerm]):
    def __init__(
        self,
        gps: Iterable[GriddedPerm],
        indices: Iterable[int],
        direction: int,
        own_col: bool = True,
        own_row: bool = True,
        ignore_parent: bool = False,
        include_empty: bool = False,
    ):
        self.gps = tuple(gps)
        self.indices = tuple(indices)
        self.direction = direction
        self.own_row, self.own_col = own_row, own_col
        self.include_empty = include_empty
        self._placed_cells = tuple(
            sorted(set(gp.pos[idx] for idx, gp in zip(self.indices, self.gps)))
        )
        possibly_empty = self.include_empty or len(self.gps) > 1
        super().__init__(ignore_parent=ignore_parent, possibly_empty=possibly_empty)

    def _placed_cell(self, idx: int) -> Cell:
        """Return the cell placed given the index of the child."""
        return self._placed_cells[idx]

    def _child_idx(self, idx: int):
        """Return the index of the child given the index of gps placed into."""
        return self._placed_cells.index(self.gps[idx].pos[self.indices[idx]])

    def placement_class(self, tiling: Tiling) -> RequirementPlacement:
        return RequirementPlacement(tiling, own_col=self.own_col, own_row=self.own_row)

    def decomposition_function(self, tiling: Tiling) -> Tuple[Tiling, ...]:
        placement_class = self.placement_class(tiling)
        placed_tilings = placement_class.place_point_of_req(
            self.gps, self.indices, self.direction
        )
        if self.include_empty:
            return (tiling.add_obstructions(self.gps),) + placed_tilings
        return placed_tilings

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if not comb_class.extra_parameters:
            return super().extra_parameters(comb_class, children)
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        algo = self.placement_class(comb_class)
        extra_parameters: Tuple[Dict[str, str], ...] = tuple({} for _ in children)
        if self.include_empty:
            child = children[0]
            for assumption in comb_class.assumptions:
                mapped_assumption = child.forward_map_assumption(assumption).avoiding(
                    child.obstructions
                )
                if mapped_assumption.gps:
                    parent_var = comb_class.get_assumption_parameter(assumption)
                    child_var = child.get_assumption_parameter(mapped_assumption)
                    extra_parameters[0][parent_var] = child_var
        for idx, (cell, child) in enumerate(
            zip(self._placed_cells, children[1:] if self.include_empty else children)
        ):
            mapped_assumptions = [
                child.forward_map_assumption(ass)
                for ass in algo.stretched_assumptions(cell)
            ]
            for assumption, mapped_assumption in zip(
                comb_class.assumptions, mapped_assumptions
            ):
                if mapped_assumption.gps:
                    parent_var = comb_class.get_assumption_parameter(assumption)
                    child_var = child.get_assumption_parameter(mapped_assumption)
                    extra_parameters[idx + 1 if self.include_empty else idx][
                        parent_var
                    ] = child_var
        return extra_parameters

    def direction_string(self):
        if self.direction == DIR_EAST:
            return "rightmost"
        if self.direction == DIR_NORTH:
            return "topmost"
        if self.direction == DIR_WEST:
            return "leftmost"
        if self.direction == DIR_SOUTH:
            return "bottommost"

    def formal_step(self):
        placing = "{}lacing the {} ".format(
            "p" if (self.own_col and self.own_row) else "partially p",
            self.direction_string(),
        )
        if len(self.gps) == 1:
            gp = self.gps[0]
            index = self.indices[0]
            if len(gp) == 1:
                return placing + "point in cell {}".format(gp.pos[index])
            if gp.is_localized():
                return placing + "{} point in {} in cell {}".format(
                    (index, gp.patt[index]), gp.patt, gp.pos[index]
                )
            return placing + "{} point in {}".format((index, gp.patt[index]), gp)
        if all(len(gp) == 1 for gp in self.gps):
            col_indices = set(x for x, _ in [gp.pos[0] for gp in self.gps])
            if len(col_indices) == 1:
                return placing + "point in column {}".format(col_indices.pop())
            row_indices = set(y for _, y in [gp.pos[0] for gp in self.gps])
            if len(row_indices) == 1:
                return placing + "point in row {}".format(row_indices.pop())
        return placing + "point at indices {} from the requirement ({})".format(
            self.indices, ", ".join(str(gp) for gp in self.gps)
        )

    def backward_cell_map(self, placed_cell: Cell, cell: Cell) -> Cell:
        x, y = cell
        if self.own_col and x > placed_cell[0] + 1:
            x -= 2
        elif self.own_col and x == placed_cell[0] + 1:
            x -= 1
        if self.own_row and y > placed_cell[1] + 1:
            y -= 2
        elif self.own_row and y == placed_cell[1] + 1:
            y -= 1
        return x, y

    def forward_gp_map(self, gp: GriddedPerm, forced_index: int) -> GriddedPerm:
        new_pos: List[Cell] = []
        forced_val = gp.patt[forced_index]
        for idx, (x, y) in enumerate(gp.pos):
            if idx == forced_index:
                if self.own_col:
                    x += 1
                if self.own_row:
                    y += 1
                new_pos.append((x, y))
            else:
                val = gp.patt[idx]
                if self.own_col and idx >= forced_index:
                    x += 2
                if self.own_row and val >= forced_val:
                    y += 2
                new_pos.append((x, y))
        return GriddedPerm(gp.patt, new_pos)

    def backward_map(
        self,
        tiling: Tiling,
        gps: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(tiling)
        idx = DisjointUnionStrategy.backward_map_index(gps)
        gp: GriddedPerm = children[idx].backward_map(cast(GriddedPerm, gps[idx]))
        if self.include_empty:
            if idx == 0:
                yield gp
                return
            idx -= 1
        placed_cell = self._placed_cell(idx)
        yield GriddedPerm(
            gp.patt, [self.backward_cell_map(placed_cell, cell) for cell in gp.pos]
        )

    def forward_map(
        self,
        tiling: Tiling,
        gp: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        indices = gp.forced_point_of_requirement(self.gps, self.indices, self.direction)
        if children is None:
            children = self.decomposition_function(tiling)
        if indices is None:
            return (children[0].forward_map(gp),) + tuple(
                None for _ in range(len(children) - 1)
            )
        gps_index, forced_index = indices
        child_index = self._child_idx(gps_index)
        if self.include_empty:
            child_index += 1
        gp = self.forward_gp_map(gp, forced_index)
        return (
            tuple(None for _ in range(child_index))
            + (children[child_index].forward_map(gp),)
            + tuple(None for _ in range(len(children) - 1))
        )

    def __str__(self) -> str:
        return "requirement placement strategy"

    def __repr__(self) -> str:
        return (
            f"RequirementPlacementStrategy(gps={self.gps}, "
            f"indices={self.indices}, direction={self.direction}, "
            f"own_col={self.own_col}, own_row={self.own_row}, "
            f"ignore_parent={self.ignore_parent}, "
            f"include_empty={self.include_empty})"
        )

    def to_jsonable(self) -> dict:
        """Return a dictionary form of the strategy."""
        d: dict = super().to_jsonable()
        d.pop("workable")
        d.pop("inferrable")
        d.pop("possibly_empty")
        d["gps"] = tuple(gp.to_jsonable() for gp in self.gps)
        d["indices"] = self.indices
        d["direction"] = self.direction
        d["own_col"] = self.own_col
        d["own_row"] = self.own_row
        d["include_empty"] = self.include_empty
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RequirementPlacementStrategy":
        gps = tuple(GriddedPerm.from_dict(gp) for gp in d.pop("gps"))
        return cls(gps=gps, **d)


class AbstractRequirementPlacementFactory(StrategyFactory[Tiling]):
    """
    Base class for requirement placement on tilings.

    It will create batch rules based on placing the direction most points
    at indices i1, ..., ik in the gridded perms g1, ..., gk, respectively.

    The point placements yielded are determined by the
    'req_indices_and_directions_to_place' function.
    """

    def __init__(
        self,
        partial: bool = False,
        ignore_parent: bool = False,
        dirs: Iterable[int] = tuple(DIRS),
        include_empty: bool = False,
    ):
        assert all(d in DIRS for d in dirs), "Got an invalid direction"
        self.partial = partial
        self.ignore_parent = ignore_parent
        self.dirs = tuple(dirs)
        self.include_empty = include_empty

    @abc.abstractmethod
    def req_indices_and_directions_to_place(
        self, tiling: Tiling
    ) -> Iterator[Tuple[Tuple[GriddedPerm, ...], Tuple[int, ...], int]]:
        """
        Iterator over all requirement lists, indices and directions to place.
        """

    def req_placements(self, tiling: Tiling) -> Tuple[RequirementPlacement, ...]:
        """
        Return the RequiremntPlacement classes used to place the points.
        """
        if self.partial:
            req_placements: Tuple[RequirementPlacement, ...] = (
                RequirementPlacement(tiling, own_row=False),
                RequirementPlacement(tiling, own_col=False),
            )
        else:
            req_placements = (RequirementPlacement(tiling),)
        return req_placements

    def __call__(self, comb_class: Tiling) -> Iterator[Rule]:
        for req_placement, (gps, indices, direction) in product(
            self.req_placements(comb_class),
            self.req_indices_and_directions_to_place(comb_class),
        ):
            if (
                direction in req_placement.directions
                and not req_placement.already_placed(gps, indices)
            ):
                strategy = RequirementPlacementStrategy(
                    gps,
                    indices,
                    direction,
                    own_row=req_placement.own_row,
                    own_col=req_placement.own_col,
                    ignore_parent=self.ignore_parent,
                    include_empty=self.include_empty,
                )
                children = req_placement.place_point_of_req(gps, indices, direction)
                if self.include_empty:
                    children = (comb_class.add_obstructions(gps),) + children
                yield strategy(comb_class, children)

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["partial"] = self.partial
        d["ignore_parent"] = self.ignore_parent
        d["dirs"] = self.dirs
        d["include_empty"] = self.include_empty
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AbstractRequirementPlacementFactory":
        return cls(**d)


class PatternPlacementFactory(AbstractRequirementPlacementFactory):
    """
    Strategy that places a single forced point of a gridded permutation.
    Yield all possible rules coming from placing a point of a pattern that
    occurs as a subpattern of requirement containing a single pattern.

    INPUTS:
        - `point_only`: only place point for length 1 subpattern.
        - `partial`: places only the point on its own row or its own column.
        - `ignore_parent`: indicate if the rule should ignore parent
        - `dirs`: The directions used for placement (default to all
          directions).
          The possible directions are:
            - `permuta.misc.DIR_NORTH`
            - `permuta.misc.DIR_SOUTH`
            - `permuta.misc.DIR_EAST`
            - `permuta.misc.DIR_WEST`
    """

    def __init__(
        self,
        point_only: bool = False,
        partial: bool = False,
        ignore_parent: bool = False,
        dirs: Iterable[int] = tuple(DIRS),
    ):
        assert all(d in DIRS for d in dirs), "Got an invalid direction"
        self.point_only = point_only
        super().__init__(partial=partial, ignore_parent=ignore_parent, dirs=dirs)

    def req_indices_and_directions_to_place(
        self, tiling: Tiling
    ) -> Iterator[Tuple[Tuple[GriddedPerm, ...], Tuple[int, ...], int]]:
        """
        If point_only, then yield all size one gps, in order to
        """
        if self.point_only:
            for cell in tiling.positive_cells:
                gps = (GriddedPerm((0,), (cell,)),)
                indices = (0,)
                for direction in self.dirs:
                    yield gps, indices, direction
        else:
            subgps = set(
                chain.from_iterable(
                    req[0].all_subperms(proper=False)
                    for req in tiling.requirements
                    if len(req) == 1
                )
            )
            for gp in subgps:
                for index, direction in product(range(len(gp)), self.dirs):
                    yield (gp,), (index,), direction

    def __str__(self) -> str:
        s = "partial " if self.partial else ""
        s += "point" if self.point_only else "requirement"
        s += " placement"
        if len(self.dirs) < 4:
            if len(self.dirs) == 1:
                s += " in direction {}".format(DIR_STR[self.dirs[0]])
            else:
                s += " in directions ".format()
                s += ", ".join(DIR_STR[d] for d in self.dirs[:-1])
                s += " and {}".format(DIR_STR[self.dirs[-1]])
        if self.ignore_parent:
            s += " (ignore parent)"
        return s

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"point_only={self.point_only}",
                f"partial={self.partial}",
                f"ignore_parent={self.ignore_parent}",
                f"dirs=({', '.join(map(DIR_REPR.__getitem__, self.dirs))},)",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d.pop("include_empty")
        d["point_only"] = self.point_only
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "PatternPlacementFactory":
        return cls(**d)


class RequirementPlacementFactory(AbstractRequirementPlacementFactory):
    """
    Strategy that places a single forced point of a requirement (a set of
    gridded permutations).
    Yield all possible rules coming from placing a point of a pattern that
    occurs as a subrequirement of a requirement on the tiling.

    INPUTS:
        - `partial`: places only the point on its own row or its own column.
        - `ignore_parent`: indicate if the rule should ignore parent
        - `dirs`: The directions used for placement (default to all
          directions).
          The possible directions are:
            - `permuta.misc.DIR_NORTH`
            - `permuta.misc.DIR_SOUTH`
            - `permuta.misc.DIR_EAST`
            - `permuta.misc.DIR_WEST`
        - `max_rules_per_req`: The limit on the size of the requirements that can
        be placed. If the product of the length of the gridded perms
        in the requirement and the number of direction is greater than the max,
        then no placements is performed for that req.
    """

    def __init__(
        self,
        subreqs: bool = False,
        partial: bool = False,
        ignore_parent: bool = False,
        dirs: Iterable[int] = tuple(DIRS),
        max_rules_per_req: Optional[int] = None,
    ):
        assert all(d in DIRS for d in dirs), "Got an invalid direction"
        self.subreqs = subreqs
        self.max_rules_per_req = max_rules_per_req
        super().__init__(partial=partial, ignore_parent=ignore_parent, dirs=dirs)

    @staticmethod
    def downward_sets(tiling: Tiling) -> Set[Tuple[GriddedPerm, ...]]:
        """Yield all requirements contained in some requirement on the tiling."""
        queue = set(tuple(sorted(req)) for req in tiling.requirements)
        # TODO: should we consider minimal gridded perms?
        #           Optimal is minimal on the factors of the tiling which
        #           contains only the requirement,
        all_reqs: Set[Tuple[GriddedPerm, ...]] = set()
        while queue:
            req = queue.pop()
            if req not in all_reqs:
                all_reqs.add(req)
                subgps = set(
                    subgp
                    for subgp in chain.from_iterable(
                        gp.all_subperms(proper=True) for gp in req
                    )
                    if len(subgp) >= 1
                )
                for subgp in subgps:
                    rest_of_req = tuple(gp for gp in req if gp.avoids(subgp))
                    newreq = tuple(sorted(rest_of_req + (subgp,)))
                    if newreq not in all_reqs:
                        queue.add(newreq)
        return all_reqs

    def req_indices_and_directions_to_place(
        self, tiling: Tiling
    ) -> Iterator[Tuple[Tuple[GriddedPerm, ...], Tuple[int, ...], int]]:
        all_reqs: Iterable[Tuple[GriddedPerm, ...]]
        if self.subreqs:
            all_reqs = self.downward_sets(tiling)
        else:
            all_reqs = tiling.requirements
        for req in all_reqs:
            if self.max_rules_per_req is not None:
                num_rule = len(self.dirs) * reduce(
                    lambda prod, req: prod * len(req), req, 1
                )
                if num_rule > self.max_rules_per_req:
                    continue
            for indices, direction in product(
                product(*[range(len(gp)) for gp in req]), self.dirs
            ):
                yield req, tuple(indices), direction

    def __str__(self) -> str:
        s = "partial " if self.partial else ""
        s += "requirement placement"
        if len(self.dirs) < 4:
            if len(self.dirs) == 1:
                s += " in direction {}".format(DIR_STR[self.dirs[0]])
            else:
                s += " in directions ".format()
                s += ", ".join(DIR_STR[d] for d in self.dirs[:-1])
                s += " and {}".format(DIR_STR[self.dirs[-1]])
        if self.ignore_parent:
            s += " (ignore parent)"
        if self.max_rules_per_req:
            s += f" (at most {self.max_rules_per_req} rule per req)"
        return s

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"subreqs={self.subreqs!r}",
                f"partial={self.partial}",
                f"ignore_parent={self.ignore_parent}",
                f"dirs=({', '.join(map(DIR_REPR.__getitem__, self.dirs))},)",
                f"max_rules_per_req={self.max_rules_per_req}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d.pop("include_empty")
        d["subreqs"] = self.subreqs
        d["max_rules_per_req"] = self.max_rules_per_req
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RequirementPlacementFactory":
        max_rules_per_req = d.pop("max_rules_per_req", None)
        return cls(max_rules_per_req=max_rules_per_req, **d)


class RowAndColumnPlacementFactory(AbstractRequirementPlacementFactory):
    def __init__(
        self,
        place_row: bool = True,
        place_col: bool = True,
        partial: bool = False,
        ignore_parent: bool = False,
        dirs: Iterable[int] = tuple(DIRS),
    ) -> None:
        assert place_col or place_row, "Must place column or row"
        self.place_row = place_row
        self.place_col = place_col
        super().__init__(
            partial=partial, ignore_parent=ignore_parent, dirs=dirs, include_empty=True
        )

    def req_indices_and_directions_to_place(
        self, tiling: Tiling
    ) -> Iterator[Tuple[Tuple[GriddedPerm, ...], Tuple[int, ...], int]]:
        """
        For each row, yield the gps with size one req for each cell in a row.
        """
        cols: Dict[int, Set[GriddedPerm]] = defaultdict(set)
        rows: Dict[int, Set[GriddedPerm]] = defaultdict(set)
        for cell in tiling.active_cells:
            gp = GriddedPerm((0,), (cell,))
            cols[cell[0]].add(gp)
            rows[cell[1]].add(gp)
        if self.place_col:
            col_dirs = tuple(d for d in self.dirs if d in (DIR_EAST, DIR_WEST))
            for gps, direction in product(cols.values(), col_dirs):
                indices = tuple(0 for _ in gps)
                yield tuple(gps), indices, direction
        if self.place_row:
            row_dirs = tuple(d for d in self.dirs if d in (DIR_NORTH, DIR_SOUTH))
            for gps, direction in product(rows.values(), row_dirs):
                indices = tuple(0 for _ in gps)
                yield tuple(gps), indices, direction

    def __str__(self) -> str:
        s = "{} placement"
        if self.place_col and self.place_row:
            s = s.format("row and column")
        elif self.place_row:
            s = s.format("row")
        else:
            s = s.format("column")
        if self.partial:
            s = " ".join(["partial", s])
        return s

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"place_row={self.place_row}",
                f"place_col={self.place_col}",
                f"partial={self.partial}",
                f"ignore_parent={self.ignore_parent}",
                f"dirs=({', '.join(map(DIR_REPR.__getitem__, self.dirs))},)",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d.pop("include_empty")
        d["place_row"] = self.place_row
        d["place_col"] = self.place_col
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RowAndColumnPlacementFactory":
        return cls(**d)


class AllPlacementsFactory(AbstractRequirementPlacementFactory):

    PLACEMENT_STRATS: Tuple[AbstractRequirementPlacementFactory, ...] = (
        PatternPlacementFactory(point_only=False),
        # subreqs=True covers everything but it blows up massively!
        RequirementPlacementFactory(subreqs=False),
        RowAndColumnPlacementFactory(place_col=True, place_row=True),
    )

    def __init__(self, ignore_parent: bool = False):
        super().__init__(ignore_parent=ignore_parent, include_empty=True)

    def req_placements(self, tiling: Tiling):
        """
        Return the RequiremntPlacement classes used to place the points.
        """
        return (
            RequirementPlacement(tiling),
            RequirementPlacement(tiling, own_row=False),
            RequirementPlacement(tiling, own_col=False),
        )

    def req_indices_and_directions_to_place(
        self, tiling: Tiling
    ) -> Iterator[Tuple[Tuple[GriddedPerm, ...], Tuple[int, ...], int]]:
        """
        Iterator over all requirement lists, indices and directions to place.
        """
        res = frozenset(
            chain.from_iterable(
                other_strat.req_indices_and_directions_to_place(tiling)
                for other_strat in AllPlacementsFactory.PLACEMENT_STRATS
            )
        )
        yield from res

    def __str__(self) -> str:
        return "all placements"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(ignore_parent={self.ignore_parent})"

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d.pop("partial")
        d.pop("dirs")
        d.pop("include_empty")
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AllPlacementsFactory":
        return cls(**d)



from collections import Counter
from functools import reduce
from operator import mul
from typing import Callable, Dict, Iterator, List, Optional, Tuple

from sympy import Eq, Function, var

from comb_spec_searcher import Constructor, Strategy
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies.constructor import DisjointUnion
from comb_spec_searcher.typing import (
    Parameters,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from tilings import GriddedPerm, Tiling


class CountComponent(Constructor[Tiling, GriddedPerm]):
    """
    The constructor used to count when we see actual components.

    The components dictionary counts how many components are to be counted by
    each parameter, noting they will no longer be tracked in the child.

    The extra_parameters map the parent parameters to the child parameters. If
    a parent parameter became empty by removing a component the parent
    parameter will not be a key in extra_parameters.
    """

    def __init__(
        self,
        parent: Tiling,
        child: Tiling,
        components: Dict[str, int],
        extra_parameters: Dict[str, str],
    ):
        self.components = components
        self.extra_parameters = extra_parameters
        self.disjoint_constructor: DisjointUnion = DisjointUnion(
            parent, (child,), (extra_parameters,)
        )
        self.indices_to_add_to = {
            parent.extra_parameters.index(k): val for k, val in components.items()
        }

    def get_equation(self, lhs_func: Function, rhs_funcs: Tuple[Function, ...]) -> Eq:
        rhs_func = rhs_funcs[0].subs(
            {b: a for a, b in self.extra_parameters.items()}, simultaneously=True
        )
        return Eq(
            lhs_func,
            rhs_func
            * reduce(mul, [var(k) ** val for k, val in self.components.items()], 1),
        )

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        terms = self.disjoint_constructor.get_terms(parent_terms, subterms, n)
        accounted_for: Terms = Counter()
        for params, value in terms.items():
            new_params = list(params)
            for idx, extra_components in self.indices_to_add_to.items():
                new_params[idx] += extra_components
            accounted_for[tuple(new_params)] += value
        return accounted_for

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]], ...]]]:
        raise NotImplementedError

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ):
        raise NotImplementedError

    @staticmethod
    def get_eq_symbol() -> str:
        return "↣"

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        raise NotImplementedError("Required for bijections")


class DetectComponentsStrategy(Strategy[Tiling, GriddedPerm]):
    @staticmethod
    def can_be_equivalent() -> bool:
        return False

    @staticmethod
    def is_two_way(comb_class: Tiling):
        return False

    @staticmethod
    def is_reversible(comb_class: Tiling):
        return False

    def shifts(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]]
    ) -> Tuple[int, ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply
        return (0,)

    @staticmethod
    def decomposition_function(comb_class: Tiling) -> Optional[Tuple[Tiling]]:
        if not comb_class.assumptions:
            return None
        return (comb_class.remove_components_from_assumptions(),)

    def constructor(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> CountComponent:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Can't detect components")

        removed_components: Dict[str, int] = {}
        for ass in comb_class.assumptions:
            value = len(ass.get_components(comb_class))
            if value:
                k = comb_class.get_assumption_parameter(ass)
                removed_components[k] = value
        return CountComponent(
            comb_class,
            children[0],
            removed_components,
            self.extra_parameters(comb_class, children)[0],
        )

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Constructor:
        raise NotImplementedError

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str]]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        extra_parameters: Dict[str, str] = {}
        child = children[0]
        for assumption in comb_class.assumptions:
            mapped_assumption = assumption.remove_components(comb_class)
            if mapped_assumption.gps:
                extra_parameters[
                    comb_class.get_assumption_parameter(assumption)
                ] = child.get_assumption_parameter(mapped_assumption)
        return (extra_parameters,)

    @staticmethod
    def formal_step() -> str:
        return "removing exact components"

    @staticmethod
    def backward_map(
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        """
        The forward direction of the underlying bijection used for object
        generation and sampling.
        """
        assert isinstance(objs[0], GriddedPerm)
        yield objs[0]

    @staticmethod
    def forward_map(
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        """
        The backward direction of the underlying bijection used for object
        generation and sampling.
        """
        return (obj,)

    @classmethod
    def from_dict(cls, d: dict) -> "DetectComponentsStrategy":
        return cls(**d)



from .assumption_insertion import AddAssumptionFactory, AddInterleavingAssumptionFactory
from .assumption_splitting import SplittingStrategy
from .detect_components import DetectComponentsStrategy
from .experimental_verification import (
    ShortObstructionVerificationStrategy,
    SubclassVerificationFactory,
)
from .factor import FactorFactory
from .fusion import ComponentFusionFactory, FusionFactory
from .obstruction_inferral import (
    EmptyCellInferralFactory,
    ObstructionInferralFactory,
    ObstructionTransitivityFactory,
    SubobstructionInferralFactory,
)
from .rearrange_assumption import RearrangeAssumptionFactory
from .requirement_insertion import (
    CellInsertionFactory,
    FactorInsertionFactory,
    RequirementCorroborationFactory,
    RequirementExtensionFactory,
    RequirementInsertionFactory,
    RootInsertionFactory,
)
from .requirement_placement import (
    AllPlacementsFactory,
    PatternPlacementFactory,
    RequirementPlacementFactory,
    RowAndColumnPlacementFactory,
)
from .row_and_col_separation import RowColumnSeparationStrategy
from .sliding import SlidingFactory
from .symmetry import SymmetriesFactory
from .verification import (
    BasicVerificationStrategy,
    DatabaseVerificationStrategy,
    ElementaryVerificationStrategy,
    InsertionEncodingVerificationStrategy,
    LocallyFactorableVerificationStrategy,
    LocalVerificationStrategy,
    MonotoneTreeVerificationStrategy,
    OneByOneVerificationStrategy,
)

__all__ = [
    # Assumptions
    "AddAssumptionFactory",
    "AddInterleavingAssumptionFactory",
    "DetectComponentsStrategy",
    "RearrangeAssumptionFactory",
    "SplittingStrategy",
    # Batch
    "CellInsertionFactory",
    "FactorInsertionFactory",
    "AllPlacementsFactory",
    "RequirementExtensionFactory",
    "RequirementInsertionFactory",
    "RequirementPlacementFactory",
    "RequirementCorroborationFactory",
    "RootInsertionFactory",
    "RowAndColumnPlacementFactory",
    # Decomposition
    "FactorFactory",
    # Equivalence
    "PatternPlacementFactory",
    "SlidingFactory",
    # Fusion
    "ComponentFusionFactory",
    "FusionFactory",
    # Inferral
    "EmptyCellInferralFactory",
    "ObstructionInferralFactory",
    "ObstructionTransitivityFactory",
    "RowColumnSeparationStrategy",
    "SubobstructionInferralFactory",
    # Symmetry
    "SymmetriesFactory",
    # Verification
    "BasicVerificationStrategy",
    "DatabaseVerificationStrategy",
    "ElementaryVerificationStrategy",
    "LocallyFactorableVerificationStrategy",
    "LocalVerificationStrategy",
    "InsertionEncodingVerificationStrategy",
    "MonotoneTreeVerificationStrategy",
    "OneByOneVerificationStrategy",
    "ShortObstructionVerificationStrategy",
    "SubclassVerificationFactory",
]



import logging
from typing import Dict, FrozenSet, Optional

from logzero import logger

from comb_spec_searcher import (
    CombinatorialSpecification,
    CombinatorialSpecificationSearcher,
)
from comb_spec_searcher.strategies.constructor import CartesianProduct, DisjointUnion
from comb_spec_searcher.strategies.rule import Rule, VerificationRule
from comb_spec_searcher.strategies.strategy import VerificationStrategy
from comb_spec_searcher.typing import CSSstrategy
from permuta import Av, Perm
from tilings import GriddedPerm, Tiling

__all__ = ["shift_from_spec"]


class TmpLoggingLevel:
    def __init__(self, level):
        self.tmp_level = level
        self.curent_level = logger.level

    def __enter__(self):
        logger.setLevel(self.tmp_level)

    def __exit__(self, exc_type, exc_value, traceback):
        logger.setLevel(self.curent_level)


class NoBasisVerification(VerificationStrategy[Tiling, GriddedPerm]):
    def __init__(self, symmetries: FrozenSet[FrozenSet[Perm]]) -> None:
        self.symmetries = symmetries
        super().__init__()

    def formal_step(self) -> str:
        cls = Av(min(self.symmetries))
        if len(self.symmetries) == 1:
            return f"No cell is {cls}"
        return f"No cell is a symmetry of {cls}"

    def verified(self, comb_class: Tiling) -> bool:
        cell_bases = (frozenset(obs) for obs, _ in comb_class.cell_basis().values())
        return not bool(self.symmetries.intersection(cell_bases))

    @classmethod
    def from_dict(cls, d: dict) -> CSSstrategy:
        raise NotImplementedError

    def to_jsonable(self) -> dict:
        raise NotImplementedError


def expanded_spec(
    tiling: Tiling, strat: VerificationStrategy, symmetries: FrozenSet[FrozenSet[Perm]]
) -> CombinatorialSpecification:
    """
    Return a spec where any tiling that does not have the basis in one cell is
    verified.
    """
    pack = strat.pack(tiling).add_verification(
        NoBasisVerification(symmetries), apply_first=True
    )
    with TmpLoggingLevel(logging.WARN):
        css = CombinatorialSpecificationSearcher(tiling, pack)
        spec = css.auto_search()
    return spec


def shift_from_spec(
    tiling: Tiling,
    strat: VerificationStrategy,
    symmetries: FrozenSet[FrozenSet[Perm]],
) -> Optional[int]:
    assert all(symmetries)
    traverse_cache: Dict[Tiling, Optional[int]] = {}
    spec = expanded_spec(tiling, strat, symmetries)

    def traverse(t: Tiling) -> Optional[int]:
        rule = spec.rules_dict[t]
        if t in traverse_cache:
            return traverse_cache[t]
        res: Optional[int]
        if isinstance(rule.strategy, NoBasisVerification):
            res = None
        elif t.dimensions == (1, 1):
            res = 0
        elif isinstance(rule, VerificationRule):
            res = shift_from_spec(tiling, rule.strategy, symmetries)
        elif isinstance(rule, Rule) and isinstance(rule.constructor, DisjointUnion):
            children_reliance = [traverse(c) for c in rule.children]
            res = min([r for r in children_reliance if r is not None], default=None)
        elif isinstance(rule, Rule) and isinstance(rule.constructor, CartesianProduct):
            min_points = [len(next(c.minimal_gridded_perms())) for c in rule.children]
            point_sum = sum(min_points)
            shifts = [point_sum - mpoint for mpoint in min_points]
            children_reliance = [traverse(c) for c in rule.children]
            res = min(
                [r + s for r, s in zip(children_reliance, shifts) if r is not None],
                default=None,
            )
        else:
            raise NotImplementedError(rule)
        traverse_cache[t] = res
        return res

    return traverse(tiling)



from typing import Dict, Iterable, Iterator, List, Optional, Sequence, Tuple, cast

from comb_spec_searcher import DisjointUnionStrategy, StrategyFactory
from comb_spec_searcher.exception import StrategyDoesNotApply
from tilings import GriddedPerm, Tiling
from tilings.algorithms import (
    AllObstructionInferral,
    ObstructionTransitivity,
    SubobstructionInferral,
)

__all__ = [
    "EmptyCellInferralFactory",
    "ObstructionInferralFactory",
    "ObstructionTransitivityFactory",
    "SubobstructionInferralFactory",
]


class ObstructionInferralStrategy(DisjointUnionStrategy[Tiling, GriddedPerm]):
    def __init__(self, gps: Iterable[GriddedPerm]):
        self.gps = tuple(sorted(gps))
        super().__init__(
            ignore_parent=True, inferrable=True, possibly_empty=False, workable=True
        )

    def decomposition_function(self, tiling: Tiling) -> Tuple[Tiling]:
        return (tiling.add_obstructions(self.gps),)

    def formal_step(self) -> str:
        """Return a string describing the operation performed."""
        if all(len(gp) == 1 for gp in self.gps):
            empty_cells_str = ", ".join(map(str, (gp.pos[0] for gp in self.gps)))
            return "the cells {{{}}} are empty".format(empty_cells_str)
        return "added the obstructions {{{}}}".format(
            ", ".join(str(p) for p in self.gps),
        )

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if not comb_class.extra_parameters:
            return super().extra_parameters(comb_class, children)
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        child = children[0]
        params: Dict[str, str] = {}
        for assumption in comb_class.assumptions:
            mapped_assumption = child.forward_map_assumption(assumption)
            if mapped_assumption.gps:
                parent_var = comb_class.get_assumption_parameter(assumption)
                child_var = child.get_assumption_parameter(mapped_assumption)
                params[parent_var] = child_var
        return (params,)

    def backward_map(
        self,
        tiling: Tiling,
        gps: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(tiling)
        yield children[0].backward_map(cast(GriddedPerm, gps[0]))

    def forward_map(
        self,
        tiling: Tiling,
        gp: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(tiling)
        return (children[0].forward_map(gp),)

    def __repr__(self) -> str:
        return self.__class__.__name__ + "(gps={})".format(self.gps)

    def __str__(self) -> str:
        return self.formal_step()

    # JSON methods

    def to_jsonable(self) -> dict:
        """Return a dictionary form of the strategy."""
        d: dict = super().to_jsonable()
        d.pop("ignore_parent")
        d.pop("inferrable")
        d.pop("possibly_empty")
        d.pop("workable")
        d["gps"] = [gp.to_jsonable() for gp in self.gps]
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "ObstructionInferralStrategy":
        gps = [GriddedPerm.from_dict(gp) for gp in d.pop("gps")]
        assert not d
        return cls(gps=gps)


class ObstructionInferralFactory(StrategyFactory[Tiling]):
    """
    A strategy used for adding obstruction that the tiling avoids, but not
    currently in the obstructions.

    Note: it isn't really a generator. This is utilised to avoid the need to
    recompute new_obs which is needed for the strategy.
    """

    def __init__(self, maxlen: int = 3):
        self.maxlen = maxlen
        super().__init__()

    def new_obs(self, tiling: Tiling) -> Sequence[GriddedPerm]:
        """
        Returns the list of new obstructions that can be added to the tiling.
        """
        return AllObstructionInferral(tiling, self.maxlen).new_obs()

    def __call__(self, comb_class: Tiling) -> Iterator[ObstructionInferralStrategy]:
        gps = self.new_obs(comb_class)
        if gps:
            yield ObstructionInferralStrategy(gps)

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["maxlen"] = self.maxlen
        return d

    @classmethod
    def from_dict(cls, d) -> "ObstructionInferralFactory":
        return cls(**d)

    def __repr__(self) -> str:
        return "{}(maxlen={})".format(self.__class__.__name__, self.maxlen)

    def __str__(self) -> str:
        if self.maxlen == 1:
            return "empty cell inferral"
        return "obstruction inferral (max length is {})".format(self.maxlen)


class EmptyCellInferralFactory(ObstructionInferralFactory):
    def __init__(self):
        super().__init__(maxlen=1)

    def to_jsonable(self):
        d = super().to_jsonable()
        d.pop("maxlen")
        return d

    @classmethod
    def from_dict(cls, d):
        return cls(**d)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


class SubobstructionInferralFactory(ObstructionInferralFactory):
    def __init__(self):
        super().__init__(maxlen=None)

    def new_obs(self, tiling: Tiling) -> List[GriddedPerm]:
        """
        Returns the list of new obstructions that can be added to the tiling.
        """
        return SubobstructionInferral(tiling).new_obs()

    def to_jsonable(self):
        d = super().to_jsonable()
        d.pop("maxlen")
        return d

    @classmethod
    def from_dict(cls, d):
        return cls(**d)

    def __repr__(self):
        return self.__class__.__name__ + "()"

    def __str__(self):
        return "subobstruction inferral"


class ObstructionTransitivityFactory(ObstructionInferralFactory):
    """
    The obstruction transitivity strategy.

    The obstruction transitivity considers all length 2 obstructions with both
    points in the same row or some column. By considering these length 2
    obstructions in similar manner as the row and column separation, as
    inequality relations. When the obstructions use a positive cell,
    transitivity applies, i.e. if a < b < c and b is positive, then a < c.
    """

    def __init__(self):
        super().__init__(maxlen=2)

    def new_obs(self, tiling: Tiling) -> Tuple[GriddedPerm, ...]:
        return ObstructionTransitivity(tiling).new_obs()

    def __str__(self) -> str:
        return "obstruction transitivity"

    def __repr__(self) -> str:
        return self.__class__.__name__ + "()"

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d.pop("maxlen")
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "ObstructionTransitivityFactory":
        assert not d, "ObstructionInferralStrategy takes no arguments"
        return cls()
